from __future__ import annotations
from typing import List, Optional
from google.protobuf import field_mask_pb2
from continual.python.sdk.environments import EnvironmentManager
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager


class ProjectManager(Manager):
    """Manages project resources."""

    name_pattern: str = "projects/{project}"

    def create(
        self,
        display_name: str,
        organization: str,
    ) -> Project:
        """Create an project.

        New projects are identified by a unique project id that is
        generated from the display name. However project ids are globally unique across
        all organizations.

        Arguments:
            display_name: Display name.
            organization: Organization resource name.

        Returns:
            A new project.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="example-org")
            >>> client.projects.create(display_name="example-proj", organization=org.name)
            <Project object {'name': 'projects/example_proj', 'display_name': 'example-proj',
            'organization': 'organizations/ceddjsa5lsrtvmkpogm0',
            'summary': {'feature_set_health': {'healthy_count': 0, 'unhealthy_count': 0, 'critical_count': 0}, '
            model_health': {'healthy_count': 0, 'unhealthy_count': 0, 'critical_count': 0}, 'feature_set_count': 0,
            'feature_set_row_count': '0', 'feature_set_bytes': '0', 'feature_count': 0, 'connection_count': 0, 'model_count': 0,
            'model_version_count': 0, 'experiment_count': '0', 'prediction_count': '0'}, 'update_time': '2022-12-15T08:25:29.713213Z',
            'create_time': '2022-12-15T08:25:29.707384Z', 'default_environment': 'projects/example_proj/environments/production'}>
        """
        req = management_pb2.CreateProjectRequest(
            project=Project(
                display_name=display_name,
                organization=organization,
            ).to_proto()
        )

        resp = self.client._management.CreateProject(req)
        return Project.from_proto(resp, client=self.client)

    def get(self, id: str) -> Project:
        """Get project.

        Arguments:
            id: Project name or id.

        Returns
            A project.

        Example:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="example-org")
            >>> proj = client.projects.create(display_name="example-proj", organization=org.name)
            >>> client.projects.get(id=proj.name).display_name
            'example-proj'
        """

        req = management_pb2.GetProjectRequest(name=self.name(id))
        resp = self.client._management.GetProject(req)
        return Project.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: Optional[str] = None,
        latest: bool = True,
    ) -> List[Project]:
        """List projects.

        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.

        Returns:
            A list of projects.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="example-org")
            >>> projs = [client.projects.create(f"project-{i}", organization=org.name) for i in range(10)]
            >>> [c.display_name for client.projects.list(page_size=10)]
            ['project-0', 'project-1', 'project-2', 'project-3', 'project-4', 'project-5', 'project-6', 'project-7', 'project-8', 'project-9']
        """

        req = management_pb2.ListProjectsRequest(
            parent=self.parent, page_size=page_size, order_by=order_by, latest=latest
        )
        resp = self.client._management.ListProjects(req)
        return [Project.from_proto(x, client=self.client) for x in resp.projects]

    def list_all(self) -> Pager[Project]:
        """List all projects.

        Pages through all projects using an iterator.

        Returns:
            A iterator of all projects.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="example-org")
            >>> projs = [client.projects.create(f"project-{i}", organization=org.name) for i in range(5)]
            >>> [c.display_name for client.projects.list_all()]
            ['project-0', 'project-1', 'project-2', 'project-3', 'project-4']
        """

        def next_page(next_page_token):
            req = management_pb2.ListProjectsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListProjects(req)
            return (
                [Project.from_proto(x, client=self.client) for x in resp.projects],
                resp.next_page_token,
            )

        return Pager(next_page)

    def delete(self, id: str) -> None:
        """Delete an project.

        Arguments:
            id: Project name or id.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="example-org")
            >>> projs = [client.projects.create(display_name=f"project-{i}", organization=org.id) for i in range(10)]
            >>> [client.projects.delete(id=proj.id)  for proj in client.projects.list_all()]
            [None, None, None, None, None, None, None, None, None, None]
            >>> len([org  for org in client.projects.list_all()])
            0
        """

        req = management_pb2.DeleteProjectRequest(name=self.name(id))
        self.client._management.DeleteProject(req)

    def update(
        self,
        id: str,
        display_name: Optional[str] = None,
    ) -> Project:
        """Update project.

        Arguments:
            id: The project's id.
            display_name:  Display name.

        Returns:
            Updated project.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="example-org")
            >>> proj = client.projects.create(display_name="name_wiht_typo", organization=org.id)
            >>> client.projects.update(id=proj.name, display_name="name_without_typo").display_name
            'name_without_typo'
        """
        paths = []
        if display_name is not None:
            paths.append("display_name")

        req = management_pb2.UpdateProjectRequest(
            update_mask=field_mask_pb2.FieldMask(paths=paths),
            project=Project(name=self.name(id), display_name=display_name).to_proto(),
        )
        resp = self.client._management.UpdateProject(req)
        return Project.from_proto(resp, client=self.client)


class Project(Resource, types.Project):
    """Project resource."""

    name_pattern: str = "projects/{project}"

    _manager: ProjectManager

    _environments: EnvironmentManager

    def _init(self):
        self._manager = ProjectManager(parent=self.parent, client=self.client)
        self._environments = EnvironmentManager(parent=self.name, client=self.client)

    @property
    def environments(self) -> EnvironmentManager:
        """Environments manager."""
        return self._environments

    def delete(self) -> None:
        """Delete project.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="example-org")
            >>> proj = client.projects.create(display_name="example-proj", organization=org.id)
            >>> proj.delete()
            >>> len([proj for proj in client.projects.list_all()])
            0
        """
        self._manager.delete(self.name)

    def update(self, display_name: Optional[str] = None) -> Project:
        """Update project.

        Arguments:
            display_name:  Display name.

        Returns:
            Updated project.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="example-org")
            >>> proj = client.projects.create(display_name="example-proj", organization=org.id)
            >>> new_proj = proj.update(display_name="updated-example-proj")
            >>> new_proj.display_name
            'updated-proj-name'
        """
        return self._manager.update(self.name, display_name=display_name)



from __future__ import annotations

from continual.python.sdk.exceptions import InvalidArgumentError


class ProjectEnvironmentIdentifer:
    """Utility class to handle project and environment names and identifers."""

    production_environment_ids = ["production"]
    default_environment_id = "production"

    def __init__(self, project_name_or_id="", environment_name_or_id=""):
        self.__project_name = ""
        self.__project_id = ""
        self.__environment_name = ""
        self.__environment_id = ""
        self.__parse(project_name_or_id, environment_name_or_id)

    def __parse(self, project_name_or_id: str, environment_name_or_id: str):
        project_part = ""
        environment_part = ""

        if project_name_or_id:
            project_part = project_name_or_id

        if environment_name_or_id:
            environment_part = environment_name_or_id
            if not project_part:
                splits = environment_part.split("/")
                if len(splits) == 4:
                    project_part = "/".join(splits[:-2])

        if not project_part:
            raise InvalidArgumentError(
                f"A project name or ID is required. Received: '{project_name_or_id}', '{environment_name_or_id}'"
            )
        elif project_part.startswith("projects/"):
            self.__project_id = project_part.split("projects/")[1]
            self.__project_name = project_part
        else:
            self.__project_id = project_part
            self.__project_name = f"projects/{self.__project_id}"

        if not environment_part:
            if self.__project_name:
                # set default as production but potentially remove
                self.__environment_id = self.default_environment_id
                self.__environment_name = (
                    f"{self.__project_name}/environments/{self.__environment_id}"
                )
            else:
                raise InvalidArgumentError(
                    f"A project name or ID is required. Received: '{project_name_or_id}', '{environment_name_or_id}'"
                )
        elif environment_part.startswith("projects/"):
            if environment_part == self.__project_name:
                self.__environment_id = self.default_environment_id
                self.__environment_name = (
                    f"{environment_part}/environments/{self.__environment_id}"
                )
            else:
                self.__environment_name = environment_part
                splits = environment_part.split("/")
                if len(splits) == 4:
                    self.__environment_id = splits[-1]
                    self.__project_name = "/".join(splits[:2])
                    self.__project_id = splits[1]
        elif environment_part in self.production_environment_ids:
            if self.__project_name:
                self.__environment_id = self.default_environment_id
                self.__environment_name = (
                    f"{self.__project_name}/environments/{self.__environment_id}"
                )
            else:
                raise InvalidArgumentError(
                    f"A project name or ID is required. Received: '{project_name_or_id}', '{environment_name_or_id}'"
                )
        else:
            self.__environment_id = environment_part
            self.__environment_name = (
                f"{self.__project_name}/environments/{self.__environment_id}"
            )

    @property
    def project_name(self) -> str:
        """Project name `projects/[project_id]`."""
        return self.__project_name

    @property
    def project_id(self) -> str:
        """Project ID `[project_id]`."""
        return self.__project_id

    @property
    def environment_name(self) -> str:
        """Environment name `projects/[project_id]/environments/[environment_id]`."""
        return self.__environment_name

    @property
    def environment_id(self) -> str:
        """Environment ID: `[environment_id]` when non-default, `[project_name]` when default."""
        return self.__environment_id

    def __getitem__(self, key):
        return getattr(self, key)



import os
import json
from pathlib import Path
from git import Repo, SymbolicReference
from git.exc import InvalidGitRepositoryError
import traceback

from continual.python.sdk.metadata import Metadata, MetadataManager
from continual.python.utils.utils import recursive_get


class ContextProvider:
    """Base context provider class."""

    name: str

    def log(self, metadata_manager: MetadataManager) -> Metadata:
        raise NotImplementedError()


class GitContext(ContextProvider):

    name: str = "git-context"

    def log(self, metadata_manager: MetadataManager) -> Metadata:
        if self._is_git_repo():
            try:
                git_context = self._get_git_metadata()
                if git_context is not None:
                    return metadata_manager.create(
                        id=self.name,
                        data=git_context,
                    )
                else:
                    return None
            except Exception as e:
                if metadata_manager.client.config.debug:
                    print(f"Error logging git context: {e}")
                    print(traceback.format_exc())
                return None

    def _is_git_repo(self) -> bool:
        try:
            _ = Repo(os.getcwd())
            return True
        except InvalidGitRepositoryError:
            return False

    def _get_git_metadata(self) -> dict:
        repo = Repo(os.getcwd(), search_parent_directories=True)
        repo_is_bare = repo.bare

        repository_name_local = ""
        if repo.working_tree_dir:
            repository_name_local = Path(repo.working_tree_dir).parts[-1]

        repository_name_remote_origin = ""
        repository_remote_origin_url = ""
        if repo.remotes and repo.remotes[0].name == "origin":
            # TODO: consider getting this from current branch tracking branch instead
            repository_name_remote_origin = os.path.splitext(
                os.path.basename(repo.remotes.origin.url)
            )[0]
            repository_remote_origin_url = repo.remotes.origin.url

        repository_name = (
            repository_name_remote_origin
            if repository_name_remote_origin
            else repository_name_local
        )

        branch_name = ""
        try:
            branch_name = repo.active_branch.name
        except:
            pass

        current_commit = None
        if not repo_is_bare:
            if isinstance(repo.head, SymbolicReference):
                current_commit = repo.head.commit
            else:
                current_commit = repo.head.reference.commit

        current_ref = None
        try:
            current_ref = repo.head.reference
        except:
            pass

        current_tag_name = ""
        current_tag_name_long = ""
        try:
            current_tag_name = repo.git.describe("--exact-match", "--abbrev=0")
            current_tag_name_long = repo.git.describe("--exact-match", "--long")
        except:
            pass

        reachable_tag_name = ""
        reachable_tag_name_long = ""
        try:
            reachable_tag_name = repo.git.describe("--abbrev=0")
            reachable_tag_name_long = repo.git.describe("--long")
        except:
            pass

        # consider setting untracked_files=True here
        is_dirty = repo.is_dirty()

        return dict(
            repository=dict(
                name=repository_name,
                name_local=repository_name_local,
                name_remote_origin=repository_name_remote_origin,
                is_bare=repo_is_bare,
                remote_origin_url=repository_remote_origin_url
                # default_branch_name
            ),
            branch=dict(name=branch_name),
            tags=dict(
                current=dict(
                    name=current_tag_name, name_describe=current_tag_name_long
                ),
                reachable=dict(
                    name=reachable_tag_name, name_describe=reachable_tag_name_long
                ),
            ),
            ref=dict(path=current_ref.path if current_ref is not None else ""),
            commit=dict(
                sha=current_commit.hexsha,
                message=current_commit.message,
                authored_datetime=current_commit.authored_datetime,
                author=dict(
                    name=current_commit.author.name, email=current_commit.author.email
                ),
                committed_datetime=current_commit.committed_datetime,
                committer=dict(
                    name=current_commit.committer.name,
                    email=current_commit.committer.email,
                ),
            )
            if current_commit is not None
            else None,
            head=dict(is_detached=repo.head.is_detached),
            is_dirty=is_dirty,
        )


class CicdRunnerContext(ContextProvider):
    name: str = "cicd-runner-context"

    def log(self, metadata_manager: MetadataManager) -> Metadata:
        if self._is_cicd_runner_context():
            try:
                cicd_runner_context = self._get_cicd_runner_metadata()
                if cicd_runner_context is not None:
                    return metadata_manager.create(
                        id=self.name,
                        data=cicd_runner_context,
                    )
                else:
                    return None
            except Exception as e:
                if metadata_manager.client.config.debug:
                    print(f"Error logging CI/CD runner context: {e}")
                    print(traceback.format_exc())
                return None

    def _is_cicd_runner_context(self) -> bool:
        # TODO: support other cicd runners
        return self._is_github_actions_cicd_runner_context()

    def _get_cicd_runner_metadata(self) -> dict:
        if self._is_github_actions_cicd_runner_context():
            return self._get_github_actions_metadata()
        return dict()

    # Github Actions

    def _is_github_actions_cicd_runner_context(self) -> bool:
        is_github_actions = False
        if os.environ.get("GITHUB_ACTIONS") == "true":
            is_github_actions = True

        return is_github_actions

    def _get_github_actions_metadata(self) -> dict:
        return dict(
            provider="Github Actions",
            run_id=f'{os.environ.get("GITHUB_RUN_ID")}-{os.environ.get("GITHUB_RUN_ATTEMPT")}',
            # Can't link to specific job because we don't have GITHUB_JOB_ID ( https://stackoverflow.com/questions/71240338/obtain-job-id-from-a-workflow-run-using-contexts )
            run_url=f'https://github.com/{os.environ.get("GITHUB_REPOSITORY")}/actions/runs/{os.environ.get("GITHUB_RUN_ID")}/attempts/{os.environ.get("GITHUB_RUN_ATTEMPT")}',
            data=dict(
                GITHUB_EVENT_NAME=os.environ.get("GITHUB_EVENT_NAME"),
                GITHUB_JOB=os.environ.get("GITHUB_JOB"),
                GITHUB_REF=os.environ.get("GITHUB_REF"),
                GITHUB_REPOSITORY=os.environ.get("GITHUB_REPOSITORY"),
                GITHUB_RUN_ID=os.environ.get("GITHUB_RUN_ID"),
                GITHUB_RUN_NUMBER=os.environ.get("GITHUB_RUN_NUMBER"),
                GITHUB_RUN_ATTEMPT=os.environ.get("GITHUB_RUN_ATTEMPT"),
                GITHUB_SHA=os.environ.get("GITHUB_SHA"),
                GITHUB_WORKFLOW=os.environ.get("GITHUB_WORKFLOW"),
                RUNNER_NAME=os.environ.get("RUNNER_NAME"),
            ),
        )


class GitProviderContext(ContextProvider):
    name: str = "git-provider-context"

    def log(self, metadata_manager: MetadataManager) -> Metadata:
        if self._is_git_provider_context():
            try:
                git_provider_context = self._get_git_provider_metadata()
                if git_provider_context is not None:
                    return metadata_manager.create(
                        id=self.name,
                        data=git_provider_context,
                    )
                else:
                    return None
            except Exception as e:
                if metadata_manager.client.config.debug:
                    print(f"Error logging git provider context: {e}")
                    print(traceback.format_exc())
                return None

    def _is_git_provider_context(self) -> bool:
        # TODO: support other providers
        return self._is_github_git_provider()

    def _get_git_provider_metadata(self) -> dict:
        if self._is_github_git_provider():
            return self._get_github_git_provider_metadata()
        return dict()

    # Github

    def _is_github_git_provider(self) -> bool:
        # TODO: Figure this out in more ways
        return self._is_github_git_provider_via_github_actions()

    def _is_github_git_provider_via_github_actions(self) -> bool:
        return CicdRunnerContext()._is_github_actions_cicd_runner_context()

    def _get_github_git_provider_metadata(self) -> dict:
        if self._is_github_git_provider_via_github_actions():
            return self._get_github_git_provider_metadata_via_github_actions()
        return dict()

    def _get_github_git_provider_metadata_via_github_actions(self) -> dict:
        github_actions_metadata = CicdRunnerContext()._get_github_actions_metadata()

        return dict(
            provider="Github",
            repository=dict(
                name=github_actions_metadata["data"]["GITHUB_REPOSITORY"],
                # In Github Actions GITHUB_REPOSITORY is `account_name/repo_name`
                url=f'https://github.com/{github_actions_metadata["data"]["GITHUB_REPOSITORY"]}',
            ),
        )


class SdlcReviewContext(ContextProvider):
    name: str = "sdlc-review-context"

    def log(self, metadata_manager: MetadataManager) -> Metadata:
        if self._is_sdlc_review_context():
            try:
                sdlc_review_context = self._get_sdlc_review_metadata()
                if sdlc_review_context is not None:
                    return metadata_manager.create(
                        id=self.name,
                        data=sdlc_review_context,
                    )
                else:
                    return None
            except Exception as e:
                if metadata_manager.client.config.debug:
                    print(f"Error logging SDLC review context: {e}")
                    print(traceback.format_exc())
                return None

    def _is_sdlc_review_context(self) -> bool:
        # TODO: support other providers
        return self._is_github_pr_context()

    def _get_sdlc_review_metadata(self) -> dict:
        if self._is_github_pr_context():
            return self._get_github_pr_metadata()
        return dict()

    # Github

    def _is_github_pr_context(self) -> bool:
        # TODO: Figure this out in more ways
        return self._is_github_pr_context_via_github_actions()

    def _is_github_pr_context_via_github_actions(self) -> bool:
        return CicdRunnerContext()._is_github_actions_cicd_runner_context()

    def _get_github_pr_metadata(self) -> dict:
        if self._is_github_pr_context_via_github_actions():
            return self._get_github_pr_metadata_via_github_actions()
        return dict()

    def _get_github_pr_metadata_via_github_actions(self) -> dict:
        github_actions_event = self._parse_github_actions_event()

        is_pr_context = recursive_get(github_actions_event, "pull_request")
        if not is_pr_context:
            return None

        return dict(
            provider="Github PR",
            id=str(recursive_get(github_actions_event, "pull_request", "number")),
            url=recursive_get(github_actions_event, "pull_request", "html_url"),
            state=recursive_get(github_actions_event, "pull_request", "state"),
            title=recursive_get(github_actions_event, "pull_request", "title"),
            user=dict(
                user_name=recursive_get(
                    github_actions_event, "pull_request", "user", "login"
                ),
                url=recursive_get(
                    github_actions_event, "pull_request", "user", "html_url"
                ),
                avatar_url=recursive_get(
                    github_actions_event, "pull_request", "user", "avatar_url"
                ),
            ),
            base=dict(
                branch=recursive_get(
                    github_actions_event, "pull_request", "base", "ref"
                ),
                sha=recursive_get(github_actions_event, "pull_request", "base", "sha"),
            ),
            head=dict(
                branch=recursive_get(
                    github_actions_event, "pull_request", "head", "ref"
                ),
                sha=recursive_get(github_actions_event, "pull_request", "head", "sha"),
            ),
            data=dict(
                # TODO
            ),
        )

    def _parse_github_actions_event(self) -> dict:
        # TODO: maybe get this from CicdRunnerContext()._get_github_actions_metadata() instead
        github_event_file = os.environ.get("GITHUB_EVENT_PATH")
        if github_event_file is None or not os.path.exists(github_event_file):
            return dict()

        try:
            with open(github_event_file, "r") as f:
                github_event = json.load(f)
        except:
            return dict()

        return github_event



import typer

from continual.python.cli import utils
from continual.python.sdk.client import Client

from rich.console import Console
from typing import List

app = typer.Typer(help="Manage models.")


def get_linked_entities(columns):
    entities = {}
    for c in columns:
        entity = c.entity
        if entity is not None and len(entity) > 0:
            entities[c.name] = entity
    return "\n".join(["%s : %s" % (k, v) for k, v in entities.items()])


def format_model_data(m, zipped=False, all_projects=False):
    create_time = m.create_time.replace(microsecond=0)
    update_time = m.update_time.replace(microsecond=0)
    num_bpj = len(m.batch_predictions.list(1000))
    num_d = len(m.promotions.list(1000))
    num_mv = len(m.model_versions.list(1000))
    latest_version = m.latest_model_version.split("/")[-1]
    if latest_version:
        mv_state = m.model_versions.get(latest_version).state.value
    else:
        mv_state = "N/A"
    current_version = m.current_version.split("/")[-1]
    mv_type = "N/A"
    mv_perf_metric = "N/A"
    mv_perf_metric_val = "N/A"
    if current_version:
        mv = m.model_versions.get(current_version)
        try:
            mv_type = mv.experiment.type.split("/")[1]
        except:
            mv_type = "N/A"
        mv_perf_metric = mv.performance_metric
        mv_perf_metric_val = round(mv.performance_metric_val, 4)
    linked_entities = get_linked_entities(m.schema.columns)
    project = m.parent.split("/")[-1]
    if zipped:
        data = [
            m.id,
            m.name,
            m.state,
            project,
            latest_version,
            mv_state,
            current_version,
            mv_type,
            mv_perf_metric,
            mv_perf_metric_val,
            m.latest_batch_prediction,
            linked_entities,
            num_mv,
            num_d,
            num_bpj,
            create_time,
            update_time,
        ]
        headers = [
            "ID",
            "Name",
            "State",
            "Project",
            "Latest Model Version",
            "Latest Version State",
            "Current Version",
            "Current Version Type",
            "Current Version Performance Metric",
            "Current Version Performance Metric Value",
            "Latest Batch Prediction",
            "Linked Entities",
            "Model Versions",
            "Promotions",
            "Batch Predictions",
            "Created",
            "Updated",
        ]
        return tuple(
            [x[0], x[1]] for x in (zip(headers, data))
        )  # for some reason list(zip) causes issues, so ...
    else:
        data = [
            m.id,
            m.state,
            latest_version,
            mv_state,
            current_version,
            mv_type,
            mv_perf_metric,
            mv_perf_metric_val,
            linked_entities,
            num_mv,
            num_d,
            num_bpj,
            create_time,
            # update_time,
        ]
        headers = [
            "ID",
            "State",
            "Latest Version",
            "Latest Version State",
            "Current Version",
            "Current Version Type",
            "Current Version Perf Metric",
            "Current Version Perf Metric Val",
            "Linked Entities",
            "Model Versions",
            "Promotions",
            "Batch Predictions",
            "Created",
            # "Updated",
        ]
        if all_projects:
            data.insert(0, m.parent.split("/")[1])
            headers.insert(0, "Project")
        return (data, headers)


# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        model_list(
            project=None,
            environment=None,
            n=30,
            filters=[],
            all_projects=False,
            style=None,
        )


@app.command("list")
@utils.exit_on_error
def model_list(
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show."),
    filters: List[str] = typer.Option([], "--filter", "-f", help="List of filters."),
    all_projects: bool = typer.Option(False, "--all", "-a", help="Show all projects."),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """List models.

    Filter can include:
        - state  (i.e. state:HEALTHY)
        - latest_mv_state (i.e. latest_mv_state:SUCCEEDED)
    """
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    filter_snippet = " (n=%s)" % n
    project_snippet = "project %s, environment %s" % (project, environment)
    if all_projects:
        project_snippet = "all accessible projects"
    if len(filters) > 0:
        filter_snippet = " with filter %s" % str(filters) + filter_snippet
    data = []
    headers = []
    for m in c.models.list(n, filters=filters, all_projects=all_projects):
        (m_data, headers) = format_model_data(m, all_projects=all_projects)
        data.append(m_data)
    typer.secho(
        "\nFound %s models in %s%s:" % (len(data), project_snippet, filter_snippet),
        fg="blue",
    )
    utils.print_table(data, headers, style=utils.get_style(style))


@app.command("get")
@utils.exit_on_error
def get(
    model: str = typer.Argument(..., help="Model ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    to_json: bool = typer.Option(
        False, "--json", help="Show full JSON representation."
    ),
):
    """Get model details."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    m = c.models.get(model)
    console = Console()
    if to_json:
        console.print(m.to_dict())
    else:
        data = format_model_data(m, zipped=True)
        typer.secho("\nRetrieving model %s: \n" % (model), fg="blue")
        utils.print_info(data)
        typer.secho("\nModel schema:\n", fg="blue")
        console.print(m.schema.to_dict())


@app.command("create-local-dev-batch-prediction")
@utils.exit_on_error
def create_local_dev(
    model: str = typer.Argument(..., help="Model ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    to_json: bool = typer.Option(
        False, "--json", help="Show full JSON representation."
    ),
):
    """Create model version for local development."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    mv = c.models.create_local_dev_model_version(model)
    console = Console()
    if to_json:
        console.print(mv.to_dict())
    else:
        typer.secho("Created model version %s \n" % (mv.name), fg="blue")


@app.command("delete")
@utils.exit_on_error
def delete(
    model: str = typer.Argument(..., help="Model ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    force: bool = typer.Option(
        False, "--force", help="Force deletion. Skips confirmation."
    ),
):
    """Delete model."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    if not force:
        force = typer.confirm(
            "Are you sure you want to delete model %s in project %s, environment %s?"
            % (model, project, environment)
        )
    if force:
        m = c.models.get(model)
        m.delete()
        typer.secho("Successfully deleted model %s." % m.id, fg="green")
    else:
        typer.secho("Cancelled model deletion!", fg="blue")



from __future__ import annotations
from typing import List, Optional, Any

from continual.python.sdk.iterators import Pager
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.rpc.management.v1 import (
    management_types_pb2,
    management_pb2,
    types as management_types_py,
)
import json
import numpy as np
import datetime


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.bool_):
            return bool(obj)
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
            return obj.isoformat()
        try:
            x = str(obj)
        except:
            pass
        else:
            return x
        return json.JSONEncoder.default(self, obj)


class MetadataManager(Manager):
    """Manages metadata resources."""

    # the name pattern for metadata depends on the resource it was created for
    name_pattern: str = ""

    def create(
        self,
        id: str = "",
        data: dict = None,
        update_if_exists: bool = True,
    ) -> Metadata:
        """Create metadata.

        Arguments:
            id: A common name used to retrieve the metadata
            data: the metadata
            update_if_exists: If true, update the metadata if it already exists

        Returns
            Metadata.

        Examples:
            >>> ... # Assume environment `env`is defined
            >>> run = env.runs.create('run0')
            >>> model_version = run.models.create("example-model").model_versions.create()
            >>> model_version.metadata.create(
            ...      id="test-map",
            ...      data={"key1": "value1", "key2": 10, "key3": 0.5},
            ... )
            <Metadata object {'name': 'projects/continual_test_proj/environments/production/models/test_model/versions/cesbtva5lsrmkt2jtbe0/metadata/test-map',
            'create_time': '2023-01-07T00:38:21.965768Z', 'update_time': '2023-01-07T00:38:21.965768Z', 'data': '{"key1": "value1", "key2": 10, "key3": 0.5}',
            'run': 'projects/continual_test_proj/environments/production/runs/cesceta5lsrkagaitff0'}>
            >>> model_version.metadata.create(
            ...     data={"key1": "value1", "key2": 10, "key3": 0.5},
            ... )
            <Metadata object {'name': 'projects/continual_test_proj/environments/production/models/test_model/versions/cesceta5lsrkagaitfh0/metadata/ceschiq5lsrkagaitfj0',
            'run': 'projects/continual_test_proj/environments/production/runs/cesceta5lsrkagaitff0', 'create_time': '2023-01-07T01:20:11.541916Z',
            'update_time': '2023-01-07T01:20:11.541916Z', 'data': '{"key1": "value1", "key2": 10, "key3": 0.5}'}>
        """

        req = management_pb2.CreateMetadataRequest(
            parent=self.parent,
            metadata=management_types_pb2.Metadata(
                run=self.run_name,
                data=json.dumps(data, cls=NpEncoder),
            ),
            metadata_id=id,
            update_if_exists=update_if_exists,
        )
        resp = self.client._management.CreateMetadata(req)
        return Metadata.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: str = None,
        latest: bool = True,
    ) -> List[Metadata]:
        """List metadata.

        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.

        Returns:
            A list of metadata.

        Examples:
            >>> ... # Assume environment `env`is defined
            >>> run = env.runs.create('run0')
            >>> model_version = run.models.create("example-model").model_versions.create()
            >>> metadatas = [model_version.metadata.create(key=f'metadata_{i}', data={'key': 'value'}) for i in range(3)]
            >>> [m.key for m in model_version.metadata.list(page_size=3)]
            ['metadata_2', 'metadata_1', 'metadata_0']
            >>> [m.key for m in model_version.metadata.list(page_size=3, latest=False)]
            ['metadata_0', 'metadata_1', 'metadata_2']
        """
        if not self.client:
            print(f"Cannot list metadata without client")
            return

        req = management_pb2.ListMetadataRequest(
            parent=self.parent,
            page_size=page_size,
            order_by=order_by,
            latest=latest,
        )
        resp = self.client._management.ListMetadata(req)
        return [Metadata.from_proto(x, client=self.client) for x in resp.metadata]

    def list_all(self) -> Pager[Metadata]:
        """List all metadata.

        Pages through all metadata using an iterator.

        Returns:
            A iterator of all metadata.

        Examples:
            >>> ... # Assume environment `env`is defined
            >>> run = env.runs.create('run0')
            >>> model_version = run.models.create("example-model").model_versions.create()
            >>> metadatas = [model_version.metadata.create(key=f'metadata_{i}', data={'key': 'value'}) for i in range(3)]
            >>> [m.key for m in model_version.metadata.list_all()]
            ['metadata_0', 'metadata_1', 'metadata_2']
        """

        def next_page(next_page_token):
            req = management_pb2.ListMetadataRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListMetadata(req)
            return (
                [Metadata.from_proto(x, client=self.client) for x in resp.metadata],
                resp.next_page_token,
            )

        return Pager(next_page)

    def get(self, id: str = "") -> Metadata:
        """Get metadata.

        Arguments:
            id: Fully qualified name or id for the metadata.

        Return
            Metadata

        Examples:
            >>> ... # Assume environment `env`is defined
            >>> run = env.runs.create('run0')
            >>> model_version = run.models.create("example-model").model_versions.create()
            >>> metadata = model_version.metadata.create(
            ...      id="test-map",
            ...      data={"key1": "value1", "key2": 10, "key3": 0.5},
            ... )
            >>> model_version.metadata.get(id=metadata.name)      # Get by name
            <Metadata object {'name': 'projects/continual_test_proj/environments/production/models/test_model/versions/cesceta5lsrkagaitfh0/metadata/ceschiq5lsrkagaitfj0',
            'run': 'projects/continual_test_proj/environments/production/runs/cesceta5lsrkagaitff0',
            'create_time': '2023-01-07T01:20:11.541916Z', 'update_time': '2023-01-07T01:20:11.541916Z',
            'data': '{"key1": "value1", "key2": 10, "key3": 0.5}'}>
            >>> model_version.metadata.get(id="test-map")         # Get by id
            <Metadata object {'name': 'projects/continual_test_proj/environments/production/models/test_model/versions/cesceta5lsrkagaitfh0/metadata/ceschiq5lsrkagaitfj0',
            'run': 'projects/continual_test_proj/environments/production/runs/cesceta5lsrkagaitff0',
            'create_time': '2023-01-07T01:20:11.541916Z', 'update_time': '2023-01-07T01:20:11.541916Z',
            'data': '{"key1": "value1", "key2": 10, "key3": 0.5}'}>
        """
        if not self.client:
            print(f"Cannot fetch metadata without client")
            return

        req = management_pb2.GetMetadataRequest(
            name=self.name(id, self.parent, f"{self.parent}/metadata/{id}")
        )
        res = self.client._management.GetMetadata(req)
        return Metadata.from_proto(res, client=self.client)

    def delete(self, id: str):
        """Delete metadata.

        Arguments:
            id: The id or fully qualified name of the metadata obj

        Examples:
            >>> ... # Assume environment `env`is defined
            >>> run = env.runs.create('run0')
            >>> model_version = run.models.create("example-model").model_versions.create()
            >>> metadata = model_version.metadata.create(
            ...      id="test-map",
            ...      data={"key1": "value1", "key2": 10, "key3": 0.5},
            ... )
            >>> len(list(model_version.metadata.list_all()))
            1
            >>> model_version.metadata.delete(name=metadata.name)
            >>> len(list(model_version.metadata.list_all()))
            0
        """
        if not self.client:
            print(f"Cannot delete metadata without client")
            return

        req = management_pb2.DeleteMetadataRequest(
            name=self.name(id, self.parent, f"{self.parent}/metadata/{id}")
        )
        self.client._management.DeleteMetadata(req)


class Metadata(Resource, management_types_py.Metadata):
    """Metadata resource."""

    # the name pattern for metadata depends on the resource it was created for
    name_pattern: str = ""

    _manager: MetadataManager
    """Metadata manager."""

    def _init(self):
        self._manager = MetadataManager(parent=self.parent, client=self.client)

    @property
    def data(self) -> str:
        return json.loads(self._attributes["data"])

    @data.setter
    def data(self, value: Any) -> None:
        if isinstance(value, str):
            management_types_py._set_attribute(
                self._attributes,
                "Metadata",
                "data",
                value,
                str,
                "",
            )
        else:
            management_types_py._set_attribute(
                self._attributes,
                "Metadata",
                "data",
                json.dumps(value, cls=NpEncoder),
                str,
                "",
            )



from __future__ import annotations
import time
from threading import Thread
from typing import List, Optional
from continual.python.sdk.artifacts import ArtifactsManager
from continual.python.sdk.batchpredictions import BatchPredictionManager
from continual.python.sdk.data_profiles import DataProfilesManager
from continual.python.sdk.dataset_version_assignments import (
    DatasetVersionAssignmentManager,
)
from continual.python.sdk.dataset_versions import DatasetVersionManager
from continual.python.sdk.events import EventManager
from continual.python.sdk.experiments import ExperimentManager
from continual.python.sdk.metrics import MetricsManager
from continual.python.sdk.model_versions import ModelVersionManager
from continual.python.sdk.models import ModelManager
from continual.python.sdk.datasets import DatasetManager
from continual.python.sdk.promotions import PromotionManager

from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.tags import TagsManager
from continual.python.sdk.metadata import MetadataManager
from continual.python.sdk.resource_checks import ResourceChecksManager
from continual.python.sdk.contexts import (
    GitContext,
    CicdRunnerContext,
    GitProviderContext,
    SdlcReviewContext,
)
from google.protobuf.timestamp_pb2 import Timestamp
from datetime import datetime


class RunManager(Manager):
    """Manages run resources."""

    name_pattern: str = "projects/{project}/environments/{environment}/runs/{run}"

    def create(
        self,
        run_id: Optional[str] = None,
        description: Optional[str] = None,
        heartbeat_interval: Optional[int] = 5,
        get_if_exists: bool = True,
    ) -> Run:
        """Create Run.

        New runs are identified by a unique run id that is
        generated from the display name.  To set a run id explicitly,
        you can pass `run_id`.  However run ids are globally unique across
        all projects.

        Arguments:
            run_id: User-defined run id.
            description: A string description of the run
            heartbeat_interval: integer number of seconds to wait between run heartbeats to
                Continual endpoint
            get_if_exists: If true, will return existing run if it exists, otherwise will raise
                an exception.

        Returns:
            A new Run.

        Examples:
            >>> ... # Assuming client, org and project is already authenticated
            >>> client.runs.create(description='Example run') # Assuming project is already set on client
            <Run object {'name': 'projects/example-proj/environments/production/runs/ceeflaq5lsrhv5c4skig',
            'description': 'Example run', 'author': 'sessions/CR6gwrc8vJuioj4iUc3VHA',
            'create_time': '2022-12-16T23:09:31.501970Z', 'update_time': '2022-12-16T23:09:31.501970Z',
            'last_heartbeat': '2022-12-16T23:09:31.517043Z', 'heartbeat_interval': '5', 'state': 'ACTIVE',
            'error_message': ''}>
            >>> proj = ... # Assume project `proj` has been created
            >>> env = proj.environments.create('example-env')
            >>> env.runs.create(description='Example run') # Can also create a run from an environment
            <Run object {'name': 'projects/example-proj/environments/example-env/runs/cxeglaq5ltrhm2c8skfa',
            'description': 'Example run', 'author': 'sessions/CR6gwrc8vJuioj4iUc3VHA',
            'create_time': '2022-12-16T23:09:31.501970Z', 'update_time': '2022-12-16T23:09:31.501970Z',
            'last_heartbeat': '2022-12-16T23:09:31.517043Z', 'heartbeat_interval': '5', 'state': 'ACTIVE',
            'error_message': ''}>
        """

        req = management_pb2.CreateRunRequest(
            run_id=run_id,
            parent=self.parent,
            description=description,
            heartbeat_interval=heartbeat_interval,
            get_if_exists=get_if_exists,
        )

        resp = self.client._management.CreateRun(req)
        run = Run.from_proto(resp, client=self.client)
        run.begin()
        return run

    def get(self, id: str) -> Run:
        """Get run.

        Arguments:
            id: Run name or id.

        Returns
            A run.

        Examples:
            >>> ... # Assuming client, org and project is already authenticated
            >>> client.runs.get('ceeflaq5lsrhv5c4skig')
            <Run object {'name': 'projects/example-proj/environments/production/runs/ceeflaq5lsrhv5c4skig',
            'description': 'Example run', 'author': 'sessions/CR6gwrc8vJuioj4iUc3VHA',
            'create_time': '2022-12-16T23:09:31.501970Z', 'update_time': '2022-12-16T23:09:31.501970Z',
            'last_heartbeat': '2022-12-16T23:09:31.517043Z', 'heartbeat_interval': '5', 'state': 'ACTIVE',
            'error_message': ''}>
        """

        req = management_pb2.GetRunRequest(name=self.name(id))
        run = self.client._management.GetRun(req)
        return Run.from_proto(run, client=self.client)

    def get_check_summary(self, id: str) -> types.RunCheckSummary:
        """Get check summary.

        Arguments:
            id: Run name or id.

        Returns
            A check summary for all resource checks made during the run id.

        Examples:
            >>> ... # Assuming client, org and project is already authenticated
            >>> client.runs.get_check_summary("ceh47hlvn3ddnpoaa77g")
            <RunCheckSummary object {'run_name': 'projects/test_project/environments/production/runs/ceh47hlvn3ddnpoaa77g', 'resource_checks': [{'name': 'projects/test_project/environments/production/datasets/test_ds/versions/ceh47hlvn3ddnpoaa7bg/resourceChecks/ceh47hlvn3ddnpoaa7hg', 'run_name': 'projects/test_project/environments/production/runs/ceh47hlvn3ddnpoaa77g', 'group_name': 'train', 'success': True, 'duration': 12.5, 'data': '{}', 'errors': ['test error'], 'warnings': ['test warning'], 'summary': 'No class imbalance detected', 'artifact_name': 'flowers.csv', 'create_time': '2022-12-20T23:22:46.708395Z', 'display_name': 'Check Class Imbalance', 'state': 'PASS', 'infos': []}, {'name': 'projects/test_project/environments/production/datasets/test_ds/versions/ceh47hlvn3ddnpoaa7bg/resourceChecks/ceh47hlvn3ddnpoaa7h0', 'run_name': 'projects/test_project/environments/production/runs/ceh47hlvn3ddnpoaa77g', 'group_name': 'train', 'duration': 15.4, 'data': '{}', 'errors': ['test error'], 'infos': ['test infos'], 'summary': 'Found duplicate indices in dataset', 'artifact_name': 'flowers.csv', 'create_time': '2022-12-20T23:22:46.698521Z', 'display_name': 'Check Duplicate Indices', 'success': False, 'state': 'PASS', 'warnings': []}, {'name': 'projects/test_project/environments/production/datasets/test_ds/versions/ceh47hlvn3ddnpoaa7bg/resourceChecks/ceh47hlvn3ddnpoaa7gg', 'run_name': 'projects/test_project/environments/production/runs/ceh47hlvn3ddnpoaa77g', 'group_name': 'test', 'success': True, 'duration': 15.4, 'data': '{}', 'warnings': ['test warning'], 'infos': ['test infos'], 'summary': 'No missing indices', 'artifact_name': 'flowers.csv', 'create_time': '2022-12-20T23:22:46.687837Z', 'display_name': 'Check Missing Index', 'state': 'PASS', 'errors': []}], 'state': 'PASS'}>
        """
        req = management_pb2.GetRunCheckSummaryRequest(name=self.name(id))
        resp = self.client._management.GetRunCheckSummary(req)
        return types.RunCheckSummary.from_proto(resp)

    def _heartbeat(self, run_name: str) -> Timestamp:
        """Send heartbeat for a given run name

        Arguments:
            run_name: string name of run

        Returns
            A Timestamp for the time of the heartbeat
        """
        req = management_pb2.LogRunHeartbeatRequest(name=run_name)
        return self.client._management.LogRunHeartbeat(req)

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: str = None,
        latest: bool = True,
    ) -> List[Run]:
        """List runs.

        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.

        Returns:
            A list of runs.

        Examples:
            >>> ... # Assuming client, org and project is already authenticated
            >>> prod_env = client.environments.get('production')
            >>> runs = [prod_env.runs.create(description=f'run{i}') for i in range(10)]
            >>> [r.description for r in prod_env.runs.list(page_size=10)]
            ['run0', 'run1', 'run2', 'run3', 'run4', 'run5', 'run6', 'run7', 'run8', 'run9']
            >>> [r.description for r in prod_env.runs.list(page_size=10, latest=False)]
            ['run9', 'run8', 'run7', 'run6', 'run5', 'run4', 'run3', 'run2', 'run1', 'run0']
        """
        req = management_pb2.ListRunsRequest(
            parent=self.parent, page_size=page_size, order_by=order_by, latest=latest
        )
        resp = self.client._management.ListRuns(req)
        return [Run.from_proto(x, client=self.client) for x in resp.runs]

    def list_all(self) -> Pager[Run]:
        """List all runs.

        Pages through all runs using an iterator.

        Returns:
            A iterator of all runs.

        Examples:
            >>> ... # Assuming client, org and project is already authenticated
            >>> prod_env = client.environments.get('production')
            >>> runs = [prod_env.runs.create(description=f'run{i}') for i in range(5)]
            >>> [r.description for r in prod_env.runs.list_all()]
            ['run0', 'run1', 'run2', 'run3', 'run4']
        """

        def next_page(next_page_token):
            req = management_pb2.ListRunsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListRuns(req)
            return (
                [Run.from_proto(x, client=self.client) for x in resp.runs],
                resp.next_page_token,
            )

        return Pager(next_page)

    def update(self, run: Run, paths: List[str]):
        """
        Arguments:
            run : Updated run object
            paths: list of actual fields to update

        Examples:
            >>> ... # Assuming client, org and project is already authenticated
            >>> prod_env = client.environments.get('production')
            >>> run = prod_env.runs.create(description='run0')
            >>> run.description = 'updated description'
            >>> prod_env.runs.update(run, ['description'])
            <Run object {'name': 'projects/test_proj_2/environments/test_env/runs/cefvjv25lsrvv8oofnl0',
            'description': 'updated description', 'author': 'sessions/dKdgYnNcYJ5tgpxd5QfQtJ',
            'create_time': '2022-12-19T05:43:24.388872Z', 'update_time': '2022-12-19T05:43:24.388872Z',
            'last_heartbeat': '2022-12-19T05:45:29.827817Z', 'heartbeat_interval': '5', 'state': 'ACTIVE',
            'error_message': ''}>
        """

        req = management_pb2.UpdateRunRequest(run=run.to_proto(), update_paths=paths)
        resp = self.client._management.UpdateRun(req)
        return Run.from_proto(resp, client=self.client)


class Run(Resource, types.Run, Thread):
    """Run resource."""

    name_pattern: str = "runs/{run}"

    _manager: RunManager

    _models: ModelManager

    _model_versions: ModelVersionManager

    _experiments: ExperimentManager

    _batch_predictions: BatchPredictionManager

    _promotions: PromotionManager

    _datasets: DatasetManager

    _dataset_versions: DatasetVersionManager

    _data_profiles: DataProfilesManager

    _assignments: DatasetVersionAssignmentManager

    _events: EventManager

    _metadata: MetadataManager

    _metrics: MetricsManager

    _tags: TagsManager

    _resource_checks: ResourceChecksManager

    _artifacts: ArtifactsManager

    def _init(self):
        Thread.__init__(self, daemon=True)
        self._manager = RunManager(parent=self.parent, client=self.client)
        self._models = ModelManager(
            parent=self.parent, client=self.client, run_name=self.name
        )
        self._datasets = DatasetManager(
            parent=self.parent, client=self.client, run_name=self.name
        )
        self._artifacts = ArtifactsManager(
            parent=self.name, client=self.client, run_name=self.name
        )

        self._tags = TagsManager(
            parent=self.name, client=self.client, run_name=self.name
        )
        self._metadata = MetadataManager(
            parent=self.name, client=self.client, run_name=self.name
        )

        self._resource_checks = ResourceChecksManager(
            parent=self.name, client=self.client, run_name=self.name
        )
        self._metrics = MetricsManager(
            parent=self.name, client=self.client, run_name=self.name
        )
        self._events = EventManager(
            parent=self.name, client=self.client, run_name=self.name
        )
        self._model_versions = ModelVersionManager(
            parent=self.name, client=self.client, run_name=self.name
        )
        self._experiments = ExperimentManager(
            parent=self.name, client=self.client, run_name=self.name
        )
        self._batch_predictions = BatchPredictionManager(
            parent=self.name, client=self.client, run_name=self.name
        )
        self._promotions = PromotionManager(
            parent=self.name, client=self.client, run_name=self.name
        )
        self._dataset_versions = DatasetVersionManager(
            parent=self.name, client=self.client, run_name=self.name
        )
        self._data_profiles = DataProfilesManager(
            parent=self.name, client=self.client, run_name=self.name
        )
        self._assignments = DatasetVersionAssignmentManager(
            parent=self.name, client=self.client, run_name=self.name
        )

    @property
    def models(self) -> ModelManager:
        """Model manager."""
        return self._models

    @property
    def model_versions(self) -> ModelVersionManager:
        """Model version manager."""
        return self._model_versions

    @property
    def experiments(self) -> ExperimentManager:
        """Experiment manager."""
        return self._experiments

    @property
    def batch_predictions(self) -> BatchPredictionManager:
        """Batch prediction manager."""
        return self._batch_predictions

    @property
    def promotions(self) -> PromotionManager:
        """Promotion manager."""
        return self._promotions

    @property
    def datasets(self) -> DatasetManager:
        """Dataset manager."""
        return self._datasets

    @property
    def dataset_versions(self) -> DatasetVersionManager:
        """Dataset version manager."""
        return self._dataset_versions

    @property
    def data_profiles(self) -> DataProfilesManager:
        """Data profiles manager."""
        return self._data_profiles

    @property
    def assignments(self) -> DatasetVersionAssignmentManager:
        """Dataset version assignment manager."""
        return self._assignments

    @property
    def events(self) -> EventManager:
        """Event manager."""
        return self._events

    @property
    def metadata(self) -> MetadataManager:
        """Metadata manager."""
        return self._metadata

    @property
    def metrics(self) -> MetricsManager:
        """Metrics manager."""
        return self._metrics

    @property
    def tags(self) -> TagsManager:
        """Tags manager."""
        return self._tags

    @property
    def resource_checks(self) -> ResourceChecksManager:
        """Resource checks manager."""
        return self._resource_checks

    @property
    def artifacts(self) -> ArtifactsManager:
        """Artifact manager."""
        return self._artifacts

    def begin(self):
        """Start Run.

        Starts the heartbeat thread for this run, and logs git context.
        """
        self._running = True
        # Thread start
        self.start()

        self._log_contexts()

    def complete(self):
        """Cleanup Run.

        This method is called when the run is completed.
        It will stop the heartbeat thread,
        set the run state to completed and set the completed time for the run.

        Examples:
            >>> ... # Assuming client, org and project is already authenticated
            >>> prod_env = client.environments.get('production')
            >>> run = prod_env.runs.create(description='run0') # Start a run
            >>> run.complete()                                 # Complete the run
        """
        self.state = "COMPLETED"
        self.complete_time = datetime.now()
        self._manager.update(self, paths=["state", "complete_time"])
        self._running = False
        self.join()

    def _heartbeat(self):
        """Sends a heartbeat to the server from this run."""
        self.last_heartbeat = self._manager._heartbeat(run_name=self.name).ToDatetime()

    def set_state(self, state: str):
        """Set run state.

        Arguments:
            state: The string name of the new state

        Examples:
            >>> ... # Assuming client, org and project is already authenticated
            >>> prod_env = client.environments.get('production')
            >>> run = prod_env.runs.create(description='run0') # Start a run
            >>> run.set_state('INACTIVE')                      # Set run state to inactive
            >>> prod_env.runs.get(run.id).state                # Get run state
            'INACTIVE'
        """
        self.state = state
        self._manager.update(self, paths=["state"])

    def get_check_summary(self):
        """Get check summary.

        Arguments:

        Returns
            A check summary for all resource checks made during this run.

        Examples:
            >>> ... # Assuming client, org and project is already authenticated
            >>> run = client.runs.get("ceh47hlvn3ddnpoaa77g")
            >>> run.get_check_summary()
            <RunCheckSummary object {'run_name': 'projects/test_project/environments/production/runs/ceh47hlvn3ddnpoaa77g', 'resource_checks': [{'name': 'projects/test_project/environments/production/datasets/test_ds/versions/ceh47hlvn3ddnpoaa7bg/resourceChecks/ceh47hlvn3ddnpoaa7hg', 'run_name': 'projects/test_project/environments/production/runs/ceh47hlvn3ddnpoaa77g', 'group_name': 'train', 'success': True, 'duration': 12.5, 'data': '{}', 'errors': ['test error'], 'warnings': ['test warning'], 'summary': 'No class imbalance detected', 'artifact_name': 'flowers.csv', 'create_time': '2022-12-20T23:22:46.708395Z', 'display_name': 'Check Class Imbalance', 'state': 'PASS', 'infos': []}, {'name': 'projects/test_project/environments/production/datasets/test_ds/versions/ceh47hlvn3ddnpoaa7bg/resourceChecks/ceh47hlvn3ddnpoaa7h0', 'run_name': 'projects/test_project/environments/production/runs/ceh47hlvn3ddnpoaa77g', 'group_name': 'train', 'duration': 15.4, 'data': '{}', 'errors': ['test error'], 'infos': ['test infos'], 'summary': 'Found duplicate indices in dataset', 'artifact_name': 'flowers.csv', 'create_time': '2022-12-20T23:22:46.698521Z', 'display_name': 'Check Duplicate Indices', 'success': False, 'state': 'PASS', 'warnings': []}, {'name': 'projects/test_project/environments/production/datasets/test_ds/versions/ceh47hlvn3ddnpoaa7bg/resourceChecks/ceh47hlvn3ddnpoaa7gg', 'run_name': 'projects/test_project/environments/production/runs/ceh47hlvn3ddnpoaa77g', 'group_name': 'test', 'success': True, 'duration': 15.4, 'data': '{}', 'warnings': ['test warning'], 'infos': ['test infos'], 'summary': 'No missing indices', 'artifact_name': 'flowers.csv', 'create_time': '2022-12-20T23:22:46.687837Z', 'display_name': 'Check Missing Index', 'state': 'PASS', 'errors': []}], 'state': 'PASS'}>
        """
        return self._manager.get_check_summary(self.name)

    def run(self):
        while True:
            if not self._running:
                break
            self._heartbeat()
            time.sleep(self.heartbeat_interval)

    def _log_contexts(self):
        """Log execution contexts as run metadata."""
        GitContext().log(self.metadata)
        GitProviderContext().log(self.metadata)
        CicdRunnerContext().log(self.metadata)
        SdlcReviewContext().log(self.metadata)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['continual',
 'continual.python',
 'continual.python.blob',
 'continual.python.cli',
 'continual.python.cli.tests',
 'continual.python.sdk',
 'continual.python.utils',
 'continual.rpc',
 'continual.rpc.graphql',
 'continual.rpc.management',
 'continual.rpc.management.v1',
 'continual.rpc.rpc',
 'continual.services']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.29',
 'click==8.0.4',
 'gitpython>=3.1.7',
 'google-cloud-storage>=1.33.0',
 'grpcio-tools>=1.43.0,<=1.48.1',
 'grpcio>=1.27.1',
 'grpcio_status>=1.31.0',
 'halo>=0.0.30',
 'humanize>=2.5.0',
 'numpy>=1.21.6',
 'omegaconf>=2.2.3',
 'pandas>=1.0.1',
 'protobuf==3.20.3',
 'pyyaml>=5.4',
 'requests>=2.28.1',
 'rich>=9.13.0',
 'scikit_learn>=1.1.2',
 'shortuuid>=1.0.9',
 'toml>=0.10.2',
 'typer==0.4.0']

entry_points = \
{'console_scripts': ['continual = continual.python.cli.cli:cli']}

setup_kwargs = {
    'name': 'continual',
    'version': '2.0.0a16',
    'description': 'Lifecyle Management for AI',
    'long_description': '# Python CLI and SDK for Continual\n\nContinual is lifecycle management for AI. Learn more at https://continual.ai.\n\n## Getting Started\n\nTo install the Continual AI CLI and SDK run:\n\n```\npip3 install continual\n```\n',
    'author': 'Continual',
    'author_email': 'support@continual.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import typer

from continual.python.cli import utils
from continual.python.sdk.client import Client

from rich.console import Console

app = typer.Typer(help="Manage runs.")


def format_run_data(r, zipped=False):
    create_time = r.create_time.replace(microsecond=0)
    update_time = r.update_time.replace(microsecond=0)
    last_heartbeat = r.last_heartbeat.replace(microsecond=0)

    if zipped:
        data = [
            r.name,
            r.description,
            r.author,
            create_time,
            update_time,
            last_heartbeat,
            r.state,
            r.error_message,
        ]
        headers = [
            "Name",
            "Description",
            "Author",
            "Created",
            "Updated",
            "Last Heartbeat",
            "State",
            "Error Message",
        ]
        return tuple([x[0], x[1]] for x in (zip(headers, data)))
    else:
        data = [
            r.name,
            r.description,
            r.author,
            create_time,
            update_time,
            last_heartbeat,
            r.state,
            r.error_message,
        ]
        headers = [
            "Name",
            "Description",
            "Author",
            "Created",
            "Updated",
            "Last Heartbeat",
            "State",
            "Error Message",
        ]
        return (data, headers)


# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        list(
            n=30,
            filters=[],
            style=None,
        )


@app.command("list")
@utils.exit_on_error
def list(
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show."),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """List Runs."""
    c = Client()
    data = []
    headers = []
    for r in c.runs.list(n):
        (r_data, headers) = format_run_data(r)
        data.append(r_data)
    typer.secho("Found %s runs:" % (len(data)), fg="blue")
    utils.print_table(data, headers, style=utils.get_style(style))


@app.command("get")
@utils.exit_on_error
def get(
    run: str = typer.Argument(..., help="Run ID."),
    json: bool = typer.Option(False, "--json", help="Show full JSON representation."),
):
    """Get run details."""
    c = Client()
    r = c.runs.get(run)
    if json:
        console = Console()
        console.print(r.to_dict())
    else:
        data = format_run_data(r, zipped=True)
        typer.secho("\nRetrieving Run %s: \n" % (run.split("/")[-1]), fg="blue")
        utils.print_info(data)


@app.command("get-check-summary")
@utils.exit_on_error
def get_check_summary(
    run: str = typer.Argument(..., help="Run ID."),
    # json: bool = typer.Option(False, "--json", help="Show full JSON representation."),
    error_on_fail: bool = typer.Option(
        False, "--error-on-fail", help="Exit with an error if check has failed."
    ),
):
    """Get check summary for run."""
    c = Client()
    r = c.runs.get_check_summary(run)
    console = Console()
    console.print(r.to_dict())

    if error_on_fail and r.state == "FAIL":
        exit(1)



from __future__ import annotations
from typing import List, Optional
import pandas as pd
import copy

from continual.python.sdk.iterators import Pager

from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager

from continual.python.sdk.utils import (
    _get_feature_analysis,
    _get_correlation_matrix,
    _get_dataset_ranges,
    _get_non_unique_index_count,
)


class DataProfilesManager(Manager):
    """Manages DataProfile resources."""

    name_pattern: str = "projects/{project}/environments/{environment}/datasets/{dataset}/dataset_versions/{version}/dataProfiles/{data_profile}"

    def create(
        self,
        dataframes: List[pd.DataFrame],
        entry_names: List[str],
        datetime_columns: List[str],
        index_column: str,
        time_index_column: str,
    ) -> DataProfile:
        """Create an data profile.

        Arguments:
            dataframes: List of dataframes on which to compute stats
            entry_names: A label for each dataset stats entry (corresponds to a dataframe)
            datetime_columns: List of columns that contain datetime values
            index_column: Name of the index column
            time_index_column: Name of the time index column

        Returns:
            A DataProfile object.

        Examples:
            >>> ... # Assume client, project, environment, and dataset are defined
            >>> import pandas as pd
            >>> num_rows = 450
            >>> data = pd.DataFrame(
            ... {
            ...     "index_col": list(range(num_rows)),
            ...     "time_index_col": pd.date_range(
            ...                         datetime.date.today(), periods=num_rows
            ...                     ).tolist(),
            ...     "timestamp_col": pd.date_range(
            ...                         datetime.date.today(), periods=num_rows, freq="2D"
            ...                     ).tolist(),
            ...     "float_col": [i * 4.0 for i in range(num_rows)],
            ...     "int_col": list(range(num_rows)),
            ...     "str_col": [f"test_val_{i}" for i in range(num_rows)],
            ...     "bool_col": [i % 2 == 0 for i in range(num_rows)],
            ...     "categorical_types": [
            ...         np.random.choice(["CAT", "DOG", "COW", "RAT"]) for _ in range(num_rows)
            ...     ],
            ...   })
            >>> dataset_version = dataset.dataset_versions.create()
            >>> dataset_version.data_profiles.create(
            ...         dataframes=[data],
            ...         entry_names=["test_entry"],
            ...         datetime_columns=["time_index_col", "timestamp_col"],
            ...         index_column="index_col",
            ...         time_index_column="time_index_col",
            ...     )
            <DataProfile object {'name': 'projects/continual_test_proj/environments/production/datasets/breast_cancer_dataset/versions/cegi8j25lsrt9r5a8ngg/data_profiles/cegi8k25lsrt9r5a8ni0',
            'dataset_stats': [{'entry_name': 'test_entry', 'num_examples': '450',
            'correlation_matrix': {'display_names': ['index_col', 'float_col', 'int_col', 'bool_col'], 'rows':...
        """

        dataset_stats = []
        for entry_name, data in zip(entry_names, dataframes):
            profile_datetime_cols = copy.deepcopy(datetime_columns)
            if time_index_column:
                profile_datetime_cols.append(time_index_column)

            results = {}
            results["entry_name"] = entry_name
            results["num_examples"] = data.shape[0]

            # only generate a correlation matri if there is more than 1 column
            if len(data.columns) > 1:
                results["correlation_matrix"] = _get_correlation_matrix(data)

            (
                results["numeric_stats"],
                results["string_stats"],
                results["categorical_stats"],
                results["timestamp_stats"],
            ) = _get_feature_analysis(data, datetime_columns, index_column=index_column)
            results["duplicate_rows"] = int(data.duplicated().sum())
            results["duplicate_indexes"] = _get_non_unique_index_count(
                data, index_col=index_column, time_index_col=time_index_column
            )

            results["dataset_ranges"] = _get_dataset_ranges(
                data=data,
                columns=profile_datetime_cols,
            )
            dataset_stats.append(results)

        req = management_pb2.CreateDataProfileRequest(
            parent=self.parent,
            data_profile=DataProfile(
                resource_name=self.parent, dataset_stats=dataset_stats
            ).to_proto(),
            run=self.run_name,
        )

        resp = self.client._management.CreateDataProfile(req)
        return DataProfile.from_proto(resp, client=self.client)

    def get(self, id: str) -> DataProfile:
        """Get data_profile.

        Arguments:
            id: DataProfile name or id.

        Returns
            A data_profile.

        Examples:
            >>> ... # Assume dataset and dataset version is defined and a data profile is created
            >>> profile = dataset_version.data_profiles.create(
            ...         dataframes=[data],
            ...         entry_names=["test_entry"],
            ...         datetime_columns=["time_index_col", "timestamp_col"],
            ...         index_column="index_col",
            ...         time_index_column="time_index_col",
            ...     )
            >>> dataset_version.data_profiles.get(profile.id)
            <DataProfile object {'name': 'projects/continual_test_proj/environments/production/datasets/breast_cancer_dataset/versions/cegi8j25lsrt9r5a8ngg/data_profiles/cegi8k25lsrt9r5a8ni0',
            'dataset_stats': [{'entry_name': 'test_entry', 'num_examples': '450',
            'correlation_matrix': {'display_names': ['index_col', 'float_col', 'int_col', 'bool_col'], 'rows':...
        """

        req = management_pb2.GetDataProfileRequest(name=self.name(id))
        data_profile = self.client._management.GetDataProfile(req)
        return DataProfile.from_proto(data_profile, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: str = None,
        latest: bool = True,
    ) -> List[DataProfile]:
        """List data profiles.

        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.

        Returns:
            A list of DataProfiles.

        Examples:
            >>> ... # Assume dataset and dataset version is defined and a data profile is created
            >>> [dataset_version.data_profiles.create(
            ...         dataframes=[data],
            ...         entry_names=["test_entry"],
            ...         datetime_columns=["time_index_col", "timestamp_col"],
            ...         index_column="index_col",
            ...         time_index_column="time_index_col",
            ...     ) for _ in range(3)]
            >>> [p.name for p in dataset_version.data_profiles.list(page_size=3))]
            [<DataProfile object {'name': 'projects/continual_test_proj/environments/production/datasets/breast_cancer_dataset/versions/cegi8j25lsrt9r5a8ngg/data_profiles/cegiasd0123ksrt9va8ni0'...
            ,<DataProfile object {'name': 'projects/continual_test_proj/environments/production/datasets/breast_cancer_dataset/versions/cegi8j25lsrt9r5a8ngg/data_profiles/cegi8k25lkjkl3t9r5a4vi0'...
            ,<DataProfile object {'name': 'projects/continual_test_proj/environments/production/datasets/breast_cancer_dataset/versions/cegi8j25lsrt9r5a8ngg/data_profiles/cegi8k25lsrt9rasd5a8ni0'...]
        """
        req = management_pb2.ListDataProfilesRequest(
            parent=self.parent,
            page_size=page_size,
            order_by=order_by,
            latest=latest,
        )
        resp = self.client._management.ListDataProfiles(req)
        return [
            DataProfile.from_proto(x, client=self.client) for x in resp.data_profiles
        ]

    def list_all(self, resource_name: str) -> Pager[DataProfile]:
        """List all data profiles.

        Pages through all data profiles using an iterator.

        Returns:
            A iterator of all DataProfiles.

        Examples:
            >>> ... # Assume dataset and dataset version is defined and a data profile is created
            >>> [dataset_version.data_profiles.create(
            ...         dataframes=[data],
            ...         entry_names=["test_entry"],
            ...         datetime_columns=["time_index_col", "timestamp_col"],
            ...         index_column="index_col",
            ...         time_index_column="time_index_col",
            ...     ) for _ in range(3)]
            >>> [p.name for p in dataset_version.data_profiles.list_all())]
            [<DataProfile object {'name': 'projects/continual_test_proj/environments/production/datasets/breast_cancer_dataset/versions/cegi8j25lsrt9r5a8ngg/data_profiles/cegiasd0123ksrt9va8ni0'...
            ,<DataProfile object {'name': 'projects/continual_test_proj/environments/production/datasets/breast_cancer_dataset/versions/cegi8j25lsrt9r5a8ngg/data_profiles/cegi8k25lkjkl3t9r5a4vi0'...
            ,<DataProfile object {'name': 'projects/continual_test_proj/environments/production/datasets/breast_cancer_dataset/versions/cegi8j25lsrt9r5a8ngg/data_profiles/cegi8k25lsrt9rasd5a8ni0'...]
        """

        def next_page(next_page_token):
            req = management_pb2.ListDataProfilesRequest(
                resource_name=resource_name, page_token=next_page_token
            )
            resp = self.client._management.ListDataProfiles(req)
            return (
                [
                    DataProfile.from_proto(x, client=self.client)
                    for x in resp.data_profiles
                ],
                resp.next_page_token,
            )

        return Pager(next_page)


class DataProfile(Resource, types.DataProfile):
    """DataProfile resource."""

    name_pattern: str = "projects/{project}/environments/{environment}/datasets/{dataset}/versions/{version}/dataProfiles/{dataProfile}"

    _manager: DataProfilesManager
    """DataProfile Manager."""

    def _init(self):
        self._manager = DataProfilesManager(parent=self.parent, client=self.client)



from __future__ import annotations
from typing import List, Optional
from continual.python.sdk.batchpredictions import (
    BatchPredictionManager,
)
from continual.python.sdk.resource_checks import ResourceChecksManager
from continual.python.sdk.data_profiles import DataProfilesManager
from continual.python.sdk.metrics import MetricsManager
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.artifacts import ArtifactsManager
from continual.python.sdk.tags import TagsManager
from continual.python.sdk.metadata import MetadataManager
from continual.python.sdk.dataset_version_assignments import (
    DatasetVersionAssignmentManager,
)


class DatasetVersionManager(Manager):
    """Manages Dataset Version resources."""

    name_pattern: str = "projects/{project}/environments/{environment}/datasets/{dataset}/versions/{version}"

    def create(self) -> DatasetVersion:
        """Create a dataset version for local development

        Returns
            A Dataset Version.

        Examples:
            >>> ... # Assume client, project, environment are defined
            >>> run = env.runs.create(description="My run")
            >>> dataset = run.datasets.get("my_dataset")
            >>> dataset_version = dataset.dataset_versions.create()
            <DatasetVersion object {'name': 'projects/test_proj_4/environments/test_env/datasets/mydataset/versions/ceg9f2i5lsrt9r5a8l8g',
            'run_name': 'projects/test_proj_4/environments/test_env/runs/ceg93ji5lsrt9r5a8kt0', 'create_time': '2022-12-19T16:55:38.879936Z',
            'update_time': '2022-12-19T16:55:38.879936Z'}>
        """
        req = management_pb2.CreateDatasetVersionRequest(
            parent=self.parent, run=self.run_name
        )
        resp = self.client._management.CreateDatasetVersion(req)
        return DatasetVersion.from_proto(resp, client=self.client)

    def get(self, id: str) -> DatasetVersion:
        """Get dataset version.

        Arguments:
            id: Dataset name or id.

        Returns
            An Dataset Version.

        Examples:
            >>> ... # Assume client, project, environment are defined
            >>> run = env.runs.create("My run")
            >>> dataset = run.datasets.get("my_dataset")
            >>> dataset_version = dataset.dataset_versions.create()
            >>> dataset.dataset_versions.get(id=dataset_version.name) # Get dataset version by name
            <DatasetVersion object {'name': 'projects/test_proj_4/environments/test_env/datasets/mydataset/versions/ceg9f2i5lsrt9r5a8l8g',
            'run_name': 'projects/test_proj_4/environments/test_env/runs/ceg93ji5lsrt9r5a8kt0', 'create_time': '2022-12-19T16:55:38.879936Z',
            'update_time': '2022-12-19T16:55:38.879936Z'}>
            >>> dataset.dataset_versions.get(id=dataset_version.id)   # Get dataset version by id
            <DatasetVersion object {'name': 'projects/test_proj_4/environments/test_env/datasets/mydataset/versions/ceg9f2i5lsrt9r5a8l8g',
            'run_name': 'projects/test_proj_4/environments/test_env/runs/ceg93ji5lsrt9r5a8kt0', 'create_time': '2022-12-19T16:55:38.879936Z',
            'update_time': '2022-12-19T16:55:38.879936Z'}>
        """
        req = management_pb2.GetDatasetVersionRequest(name=self.name(id))
        resp = self.client._management.GetDatasetVersion(req)
        return DatasetVersion.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: str = None,
        latest: bool = True,
    ) -> List[DatasetVersion]:
        """List dataset versions.

        Arguments:
            page_size: Number of items to return.

        Returns:
            A list of dataset versions.

        Examples:
            >>> ... # Assume client, project, environment are defined
            >>> run = env.runs.create("My run")
            >>> dataset = run.datasets.get("my_dataset")
            >>> dvs = [dataset.dataset_versions.create() for _ in range(3)]
            >>> first_dataset_version = dataset.dataset_versions.list(page_size=10)[0]               # Get first dataset version
            >>> latest_dataset_version = dataset.dataset_versions.list(page_size=10, latest=True)[0] # Get latest dataset version
        """
        req = management_pb2.ListDatasetVersionsRequest(
            parent=self.parent,
            page_size=page_size,
            order_by=order_by,
            latest=latest,
        )
        resp = self.client._management.ListDatasetVersions(req)
        return [
            DatasetVersion.from_proto(x, client=self.client)
            for x in resp.dataset_versions
        ]

    def list_all(self) -> Pager[DatasetVersion]:
        """List all dataset versions.

        Pages through all dataset versions using an iterator.

        Returns:
            A iterator of all dataset versions.

        Examples:
            >>> ... # Assume client, project, environment are defined
            >>> run = env.runs.create("My run")
            >>> dataset = run.datasets.get("my_dataset")
            >>> dvs = [dataset.dataset_versions.create() for _ in range(3)]
            >>> len(list(dataset.dataset_versions.list_all())) # List all dataset versions
            3
        """

        def next_page(next_page_token):
            req = management_pb2.ListDatasetVersionsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListDatasetVersions(req)
            return (
                [
                    DatasetVersion.from_proto(x, client=self.client)
                    for x in resp.dataset_versions
                ],
                resp.next_page_token,
            )

        return Pager(next_page)


class DatasetVersion(Resource, types.DatasetVersion):
    """Dataset version resource."""

    name_pattern: str = "projects/{project}/environments/{environment}/datasets/{dataset}/versions/{version}"
    _manager: DatasetVersionManager
    """Dataset version manager."""

    _resource_checks: ResourceChecksManager
    """Resource Checks Manager"""

    _data_profiles: DataProfilesManager
    """Data Profiles Manager"""

    _assignments: DatasetVersionAssignmentManager
    """Dataset Version Assignment Manager"""

    _metrics: MetricsManager
    """Metrics Manager"""

    _artifacts: ArtifactsManager
    """Artifacts Manager"""

    _tags: TagsManager
    """Tags Manager"""

    _metadata: MetadataManager
    """Metadata Manager"""

    def _init(self):
        self._manager = DatasetVersionManager(
            parent=self.parent, client=self.client, run_name=self.run
        )
        self._artifacts = ArtifactsManager(
            parent=self.name, client=self.client, run_name=self.run
        )
        self._tags = TagsManager(
            parent=self.name, client=self.client, run_name=self.run
        )
        self._metadata = MetadataManager(
            parent=self.name, client=self.client, run_name=self.run
        )
        self._resource_checks = ResourceChecksManager(
            parent=self.name, client=self.client, run_name=self.run
        )
        self._data_profiles = DataProfilesManager(
            parent=self.name, client=self.client, run_name=self.run
        )
        self._batch_predictions = BatchPredictionManager(
            parent=self.parent, client=self.client, run_name=self.run
        )
        self._metrics = MetricsManager(
            parent=self.name, client=self.client, run_name=self.run
        )
        self._assignments = DatasetVersionAssignmentManager(
            parent=self.name, client=self.client, run_name=self.run
        )

    @property
    def artifacts(self) -> ArtifactsManager:
        """Artifacts manager."""
        return self._artifacts

    @property
    def tags(self) -> TagsManager:
        """Tags manager."""
        return self._tags

    @property
    def metadata(self) -> MetadataManager:
        """Metadata manager."""
        return self._metadata

    @property
    def resource_checks(self) -> ResourceChecksManager:
        """Resource Checks manager."""
        return self._resource_checks

    @property
    def data_profiles(self) -> DataProfilesManager:
        """Data Profiles manager."""
        return self._data_profiles

    @property
    def batch_predictions(self) -> BatchPredictionManager:
        """Batch Predictions manager."""
        return self._batch_predictions

    @property
    def metrics(self) -> MetricsManager:
        """Metrics manager."""
        return self._metrics

    @property
    def assignments(self) -> DatasetVersionAssignmentManager:
        """Dataset Version Assignment manager."""
        return self._assignments



from __future__ import annotations
from typing import Iterator, List, Optional
from continual.python.sdk.metrics import MetricsManager

from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.artifacts import ArtifactsManager
from continual.python.sdk.resource_checks import ResourceChecksManager
from continual.python.sdk.tags import TagsManager
from continual.python.sdk.metadata import MetadataManager


class BatchPredictionManager(Manager):
    """Manages Batch Prediction resources."""

    name_pattern: str = "projects/{project}/environments/{environment}/models/{model}/batchPredictions/{batch_prediction_job}"

    def create(self, model_version_name: str) -> BatchPrediction:
        """Get batch prediction job.

        Arguments:
            model_version_name: Name of the model version to use for prediction

        Returns
            A Batch prediction.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> model.batch_predictions.create(model_version_name=model_version.name)     # Create from specific model version
            <BatchPrediction object {'name': 'projects/test_proj_4/environments/test_env/models/my_model/batchPredictions/ceg9bk25lsrt9r5a8l5g',
            'state': 'CREATED', 'model_version': 'projects/test_proj_4/environments/test_env/models/my_model/versions/ceg99oq5lsrt9r5a8l2g',
            'create_time': '2022-12-19T16:48:16.368146Z', 'run_name': 'projects/test_proj_4/environments/test_env/runs/ceg93ji5lsrt9r5a8kt0',
            'prediction_count': '0', 'error_message': '', 'stack_trace': ''}>
        """
        req = management_pb2.CreateBatchPredictionRequest(
            parent=self.parent,
            run_name=self.run_name,
            trained_model_version_name=model_version_name,
        )
        resp = self.client._management.CreateBatchPrediction(req)
        return BatchPrediction.from_proto(resp, client=self.client)

    def get(self, id: str) -> BatchPrediction:
        """Get batch prediction job.

        Arguments:
            id: Batch Prediction  name or id.

        Returns
            A Batch prediction.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> batchprediction = model.batch_predictions.create(model_version_name=model_version.name)
            >>> model.batch_predictions.get(batchprediction.id)     # Get from id
            <BatchPrediction object {'name': 'projects/test_proj_4/environments/test_env/models/my_model/batchPredictions/ceg9bk25lsrt9r5a8l5g',
            'state': 'CREATED', 'model_version': 'projects/test_proj_4/environments/test_env/models/my_model/versions/ceg99oq5lsrt9r5a8l2g',
            'create_time': '2022-12-19T16:48:16.368146Z', 'run_name': 'projects/test_proj_4/environments/test_env/runs/ceg93ji5lsrt9r5a8kt0',
            'prediction_count': '0', 'error_message': '', 'stack_trace': ''}>
        """
        req = management_pb2.BatchPredictionRequest(name=self.name(id))
        resp = self.client._management.GetBatchPrediction(req)
        return BatchPrediction.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        latest: bool = True,
        order_by: Optional[str] = None,
        all_projects: bool = False,
    ) -> List[BatchPrediction]:
        """List batch prediction jobs.

        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.
            all_projects: Whether to include all instances of this resource from the project or just from the current parent.

        Returns:
            A list of batch prediction jobs.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> batchpredictions = [model.batch_predictions.create(model_version_name=model_version.name) for _ in range(10)]
            >>> len(model.batch_predictions.list(page_size=10))     # List 10 batch predictions
            10
        """
        req = management_pb2.ListBatchPredictionsRequest(
            parent=self.parent,
            latest=latest,
            order_by=order_by,
            page_size=page_size,
            all_projects=all_projects,
        )
        resp = self.client._management.ListBatchPredictions(req)
        return [
            BatchPrediction.from_proto(u, client=self.client)
            for u in resp.batch_predictions
        ]

    def list_all(self) -> Iterator[BatchPrediction]:
        """List all batch prediction jobs.

        Pages through all batch prediction jobs using an iterator.

        Returns:
            A iterator of all batch prediction jobs.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> batchpredictions = [model.batch_predictions.create(model_version_name=model_version.name) for _ in range(10)]
            >>> len(list(model.batch_predictions.list_all()))     # List all batch predictions
            10
        """

        def next_page(next_page_token):
            req = management_pb2.ListBatchPredictionsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListBatchPredictions(req)
            return (
                [
                    BatchPrediction.from_proto(u, client=self.client)
                    for u in resp.batch_predictions
                ],
                resp.next_page_token,
            )

        return Pager(next_page)


class BatchPrediction(Resource, types.BatchPrediction):
    """BatchPrediction resource."""

    name_pattern: str = "projects/{project}/environments/{environment}/models/{model}/batchPredictions/{batch_prediction_job}"

    _manager: BatchPredictionManager
    """BatchPrediction manager"""

    _metrics: MetricsManager
    """Metrics Manager"""

    _artifacts: ArtifactsManager
    """Artifacts Manager"""

    _tags: TagsManager
    """Tags Manager"""

    _metadata: MetadataManager
    """Metadata Manager"""

    _resource_checks: ResourceChecksManager

    def _init(self):
        self._manager = BatchPredictionManager(parent=self.parent, client=self.client)
        self._artifacts = ArtifactsManager(
            parent=self.name, client=self.client, run_name=self.run_name
        )
        self._tags = TagsManager(parent=self.name, client=self.client)
        self._metadata = MetadataManager(parent=self.name, client=self.client)
        self._metrics = MetricsManager(
            parent=self.name, client=self.client, run_name=self.run_name
        )
        self._resource_checks = ResourceChecksManager(
            parent=self.name, client=self.client, run_name=self.run_name
        )

    @property
    def artifacts(self) -> ArtifactsManager:
        """Artifacts Manager."""
        return self._artifacts

    @property
    def tags(self) -> TagsManager:
        """Tags Manager."""
        return self._tags

    @property
    def metadata(self) -> MetadataManager:
        """Metadata Manager."""
        return self._metadata

    @property
    def metrics(self) -> MetricsManager:
        """Metrics Manager."""
        return self._metrics

    @property
    def resource_checks(self) -> ResourceChecksManager:
        """Resource Checks manager."""
        return self._resource_checks



from __future__ import annotations

from typing import List, Optional
from continual.python.sdk.batchpredictions import (
    BatchPredictionManager,
)
from continual.python.sdk.events import EventManager
from continual.python.sdk.resource_checks import ResourceChecksManager
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.experiments import ExperimentManager
from continual.python.sdk.promotions import PromotionManager
from continual.python.sdk.metrics import MetricsManager
from continual.python.sdk.artifacts import ArtifactsManager
from continual.python.sdk.tags import TagsManager
from continual.python.sdk.metadata import MetadataManager


class ModelVersionManager(Manager):
    """Manages Model Version resources."""

    name_pattern: str = "projects/{project}/environments/{environment}/models/{model}/versions/{version}"

    def create(self) -> ModelVersion:
        """Create a model version for local development

        Returns
            A Model Version.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model.model_versions.create()
            <ModelVersion object {'name': 'projects/test_proj_4/environments/test_env/models/my_model/versions/ceg98ea5lsrt9r5a8l10',
            'run_name': 'projects/test_proj_4/environments/test_env/runs/ceg93ji5lsrt9r5a8kt0',
            'state': 'CREATED', 'create_time': '2022-12-19T16:41:29.232614Z', 'update_time': '2022-12-19T16:41:29.232614Z',
            'experiment_name': '', 'error_message': '', 'stack_trace': '', 'training_row_count': '0', 'validation_row_count': '0',
            'test_row_count': '0', 'performance_metric': '', 'performance_metric_val': 0.0, 'promotion': '', 'promoted': False}>
        """
        req = management_pb2.CreateModelVersionRequest(
            parent=self.parent, run_name=self.run_name
        )
        resp = self.client._management.CreateModelVersion(req)
        return ModelVersion.from_proto(resp, client=self.client)

    def get(self, id: str) -> ModelVersion:
        """Get model version.

        Arguments:
            id: Model name or id.

        Returns
            An Model Version.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> model.model_versions.get(model_version.id)
            <ModelVersion object {'name': 'projects/test_proj_4/environments/test_env/models/my_model/versions/ceg98ea5lsrt9r5a8l10',
            'run_name': 'projects/test_proj_4/environments/test_env/runs/ceg93ji5lsrt9r5a8kt0',
            'state': 'CREATED', 'create_time': '2022-12-19T16:41:29.232614Z', 'update_time': '2022-12-19T16:41:29.232614Z',
            'experiment_name': '', 'error_message': '', 'stack_trace': '', 'training_row_count': '0', 'validation_row_count': '0',
            'test_row_count': '0', 'performance_metric': '', 'performance_metric_val': 0.0, 'promotion': '', 'promoted': False}>
        """
        req = management_pb2.GetModelVersionRequest(name=self.name(id))
        resp = self.client._management.GetModelVersion(req)
        return ModelVersion.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: str = None,
        latest: bool = True,
        all_projects: bool = False,
    ) -> List[ModelVersion]:
        """List model versions.

        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.
            all_projects: Whether to include all instances of this resource from the project or just from the current parent.

        Returns:
            A list of model versions.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> models = [model.model_versions.create() for _ in range(3)]
            >>> first_model_version = model.model_versions.list(page_size=10)[0]               # Get first model version
            >>> latest_model_version = model.model_versions.list(page_size=10, latest=True)[0] # Get latest model version
        """
        req = management_pb2.ListModelVersionsRequest(
            parent=self.parent,
            page_size=page_size,
            all_projects=all_projects,
            order_by=order_by,
            latest=latest,
        )
        resp = self.client._management.ListModelVersions(req)
        return [
            ModelVersion.from_proto(x, client=self.client) for x in resp.model_versions
        ]

    def list_all(self) -> Pager[ModelVersion]:
        """List all model versions.

        Pages through all model versions using an iterator.

        Returns:
            A iterator of all model versions.

        Examples:
            >>> ... # Assume client, project, environment are defined
            >>> run = env.runs.create("My run")
            >>> model = run.models.get("my_model")
            >>> dvs = [model.model_versions.create() for _ in range(3)]
            >>> len(list(model.model_versions.list_all())) # List all model versions
            3
        """

        def next_page(next_page_token):
            req = management_pb2.ListModelVersionsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListModelVersions(req)
            return (
                [
                    ModelVersion.from_proto(x, client=self.client)
                    for x in resp.model_versions
                ],
                resp.next_page_token,
            )

        return Pager(next_page)


class ModelVersion(Resource, types.ModelVersion):
    """Model version resource."""

    name_pattern: str = "projects/{project}/environments/{environment}/models/{model}/versions/{version}"

    _manager: ModelVersionManager

    _experiments: ExperimentManager

    _events: EventManager

    _promotions: PromotionManager

    _resource_checks: ResourceChecksManager

    _batch_predictions: BatchPredictionManager

    _metrics: MetricsManager

    _artifacts: ArtifactsManager

    _tags: TagsManager

    _metadata: MetadataManager

    def _init(self):
        self._manager = ModelVersionManager(
            parent=self.parent, client=self.client, run_name=self.run_name
        )
        self._promotions = PromotionManager(
            parent=self.parent, client=self.client, run_name=self.run_name
        )
        self._batch_predictions = BatchPredictionManager(
            parent=self.parent, client=self.client, run_name=self.run_name
        )
        self._experiments = ExperimentManager(
            parent=self.name, client=self.client, run_name=self.run_name
        )
        self._events = EventManager(
            parent=self.name, client=self.client, run_name=self.run_name
        )
        self._metrics = MetricsManager(
            parent=self.name, client=self.client, run_name=self.run_name
        )
        self._artifacts = ArtifactsManager(
            parent=self.name, client=self.client, run_name=self.run_name
        )
        self._tags = TagsManager(
            parent=self.name, client=self.client, run_name=self.run_name
        )
        self._metadata = MetadataManager(
            parent=self.name, client=self.client, run_name=self.run_name
        )

        self._resource_checks = ResourceChecksManager(
            parent=self.name, client=self.client, run_name=self.run_name
        )

    @property
    def promotions(self) -> PromotionManager:
        """Promotion manager."""
        return self._promotions

    @property
    def batch_predictions(self) -> BatchPredictionManager:
        """Batch Prediction manager."""
        return self._batch_predictions

    @property
    def experiments(self) -> ExperimentManager:
        """Experiment manager."""
        return self._experiments

    @property
    def metrics(self) -> MetricsManager:
        """Metrics manager."""
        return self._metrics

    @property
    def artifacts(self) -> ArtifactsManager:
        """Artifacts manager."""
        return self._artifacts

    @property
    def tags(self) -> TagsManager:
        """Tags manager."""
        return self._tags

    @property
    def resource_checks(self) -> ResourceChecksManager:
        """Resource Checks manager."""
        return self._resource_checks

    @property
    def metadata(self) -> MetadataManager:
        """Metadata manager."""
        return self._metadata

    @property
    def events(self) -> EventManager:
        """Event manager."""
        return self._events



from __future__ import annotations
from continual.python.sdk.artifacts import ArtifactsManager
from continual.python.sdk.batchpredictions import BatchPredictionManager
from continual.python.sdk.data_profiles import DataProfilesManager
from continual.python.sdk.dataset_version_assignments import (
    DatasetVersionAssignmentManager,
)
from continual.python.sdk.dataset_versions import DatasetVersionManager
from continual.python.sdk.datasets import DatasetManager
from continual.python.sdk.events import EventManager
from continual.python.sdk.experiments import ExperimentManager
from continual.python.sdk.metadata import MetadataManager
from continual.python.sdk.metrics import MetricsManager
from continual.python.sdk.model_versions import ModelVersionManager
from continual.python.sdk.models import ModelManager
from continual.python.sdk.promotions import PromotionManager
from continual.python.sdk.resource_checks import ResourceChecksManager
from continual.python.sdk.tags import TagsManager
import grpc
from typing import Optional, Tuple
from google.protobuf.empty_pb2 import Empty
from continual.python.sdk.config import Config
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import management_pb2_grpc
from continual.python.sdk.projects import ProjectManager, Project
from continual.python.sdk.organizations import OrganizationManager
from continual.python.sdk.runs import RunManager
from continual.python.sdk.users import UserManager, User
from continual.python.sdk.interceptors import AuthInterceptor
from continual.python.sdk.exceptions import normalize_exceptions_for_class
from continual.python.utils.client_utils import get_management_channel
from importlib.metadata import version

try:
    __version__ = version("continual")
except:
    __version__ = "local-dev"
from continual.python.sdk.identifiers import ProjectEnvironmentIdentifer


class Client:
    """Continual client."""

    _users: UserManager

    _projects: ProjectManager

    _organizations: OrganizationManager

    _runs: RunManager

    _models: ModelManager

    _model_versions: ModelVersionManager

    _experiments: ExperimentManager

    _batch_predictions: BatchPredictionManager

    _promotions: PromotionManager

    _datasets: DatasetManager

    _dataset_versions: DatasetVersionManager

    _data_profiles: DataProfilesManager

    _assignments: DatasetVersionAssignmentManager

    _events: EventManager

    _metadata: MetadataManager

    _metrics: MetricsManager

    _tags: TagsManager

    _resource_checks: ResourceChecksManager

    _artifacts: ArtifactsManager

    _config: Config

    _management: management_pb2_grpc.ManagementAPIStub

    def __init__(
        self,
        api_key: Optional[str] = None,
        endpoint: Optional[str] = None,
        project: Optional[str] = None,
        environment: Optional[str] = None,
        verify: bool = True,
    ):
        """Initialize client.

        It is recommended to use `continual login` to generate a
        local on-disk API key.

        Arguments:
            api_key: API key.
            project: Default project.
            endpoint: Continual endpoint. Default is http://sdk.continual.ai.
            environment: Project environment
            verify: Whether or not to verify the arguments on init

        Examples:
            >>> from continual import Client
            >>> c = Client(verify=False)   # Does not verify API key and project name
            <continual.python.sdk.client.Client object at 0x7f63a3e2b4c0>
            >>> c = Client(endpoint="http://sdk.continual.ai", api_key="sessions/l123kjnal34AiS2", project="example-project", environment="example-env", verify=False)
            <continual.python.sdk.client.Client object at 0x7f63a3e2b4c0>
            >>> c.config.endpoint, c.config.project, c.config.environment
            ('http://sdk.continual.ai', 'projects/example-project', 'example-env')
            >>> Client(verify=True)  # Will attempt to verify API key, project name, and environment in YAML or ENV config
        """
        # Initialize config to process args
        self._config = Config(
            endpoint=endpoint, api_key=api_key, project=project, environment=environment
        )

        if self.config.api_key and verify:
            self.set_config_api_key(api_key=self.config.api_key, save=False)
        else:
            # At the very least initialie the management channel to allow register and login
            self._init_grpc_connnections(api_key=self.config.api_key)
            self._init_managers()

        # Initialize managers that only depend on client
        self._users = UserManager(client=self)
        self._organizations = OrganizationManager(client=self)
        self._projects = ProjectManager(client=self)

        # Verify and set proj and env IF they are set in config
        if self.config.project and verify:
            self.set_config_project(
                project=self.config.project,
                save=False,
                environment=self.config.environment,
            )
            if self.config.environment is None or self.config.environment == "":
                self.set_config_environment(self.config.environment, save=False)

        if environment is not None and verify:
            self.set_config_environment(environment, save=False)

    @property
    def users(self) -> UserManager:
        """User manager."""
        return self._users

    @property
    def projects(self) -> ProjectManager:
        """Project manager."""
        return self._projects

    @property
    def organizations(self) -> OrganizationManager:
        """Organization manager."""
        return self._organizations

    @property
    def runs(self) -> RunManager:
        """Run manager."""
        return self._runs

    @property
    def models(self) -> ModelManager:
        """Model manager."""
        return self._models

    @property
    def model_versions(self) -> ModelVersionManager:
        """Model version manager."""
        return self._model_versions

    @property
    def experiments(self) -> ExperimentManager:
        """Experiment manager."""
        return self._experiments

    @property
    def batch_predictions(self) -> BatchPredictionManager:
        """Batch prediction manager."""
        return self._batch_predictions

    @property
    def promotions(self) -> PromotionManager:
        """Promotion manager."""
        return self._promotions

    @property
    def datasets(self) -> DatasetManager:
        """Dataset manager."""
        return self._datasets

    @property
    def dataset_versions(self) -> DatasetVersionManager:
        """Dataset version manager."""
        return self._dataset_versions

    @property
    def data_profiles(self) -> DataProfilesManager:
        """Data profiles manager."""
        return self._data_profiles

    @property
    def assignments(self) -> DatasetVersionAssignmentManager:
        """Dataset version assignment manager."""
        return self._assignments

    @property
    def events(self) -> EventManager:
        """Event manager."""
        return self._events

    @property
    def metadata(self) -> MetadataManager:
        """Metadata manager."""
        return self._metadata

    @property
    def metrics(self) -> MetricsManager:
        """Metrics manager."""
        return self._metrics

    @property
    def tags(self) -> TagsManager:
        """Tags manager."""
        return self._tags

    @property
    def resource_checks(self) -> ResourceChecksManager:
        """Resource checks manager."""
        return self._resource_checks

    @property
    def artifacts(self) -> ArtifactsManager:
        """Artifact manager."""
        return self._artifacts

    @property
    def config(self) -> Config:
        """User manager."""
        return self._config

    def _init_managers(self):
        # Initialize managers.
        parent = self.config.environment
        if parent is None:
            parent = "projects/-"
        else:
            env = self.config.environment
            if (
                env == None
                or env == "master"
                or env == "main"
                or env == "production"
                or len(env) == 0
            ):
                env = "production"

            splits = self.config.environment.split("/")
            if (
                len(splits) != 4
                or splits[0] != "projects"
                or splits[2] != "environments"
            ):
                parent = f"{self.config.project}/environments/{self.config.environment}"

        self._runs = RunManager(client=self, parent=parent)
        self._models = ModelManager(client=self, parent=parent)
        self._model_versions = ModelVersionManager(client=self, parent=parent)
        self._experiments = ExperimentManager(client=self, parent=parent)
        self._batch_predictions = BatchPredictionManager(client=self, parent=parent)
        self._promotions = PromotionManager(client=self, parent=parent)
        self._datasets = DatasetManager(client=self, parent=parent)
        self._dataset_versions = DatasetVersionManager(client=self, parent=parent)
        self._data_profiles = DataProfilesManager(client=self, parent=parent)
        self._assignments = DatasetVersionAssignmentManager(client=self, parent=parent)
        self._events = EventManager(client=self, parent=parent)
        self._metadata = MetadataManager(client=self, parent=parent)
        self._metrics = MetricsManager(client=self, parent=parent)
        self._tags = TagsManager(client=self, parent=parent)
        self._resource_checks = ResourceChecksManager(client=self, parent=parent)
        self._artifacts = ArtifactsManager(client=self, parent=parent)

    def _init_grpc_connnections(self, api_key):
        """Init auth"""
        auth_interceptor = AuthInterceptor(lambda: api_key)
        self._mgmt_channel = get_management_channel(self.config.endpoint)

        self._mgmt_channel = grpc.intercept_channel(
            self._mgmt_channel, auth_interceptor
        )
        self._management = normalize_exceptions_for_class(
            management_pb2_grpc.ManagementAPIStub(self._mgmt_channel)
        )

    def _verify_api_key(self):
        """Verifies that the API key is valid by calling CheckViewer"""
        try:
            self.check_viewer()
        except Exception as e:
            api_key = self._mgmt_channel._interceptor.api_key_getter()
            raise Exception(f"Unable to verify API key {api_key}. {str(e)}")

    def set_config_api_key(self, api_key: str, save: bool = False) -> None:
        """Sets config API key.

        Arguments:
            api_key: The API key to set on this client.
            save: Whether to save the API key to the config file on disk.

        Examples:
            >>> from continual import Client
            >>> c = Client(verify=False)
            >>> c.config.api_key # If a YAML config is present in the OS env or filesystem this will not be empty
            ''
            >>> c.set_config_api_key(api_key='sessions/aJcVPD9Vh6q3ofYkuWyzQS', save=False)
            >>> c.config.api_key
            'sessions/aJcVPD9Vh6q3ofYkuWyzQS'

        """
        # Reinitalize Initialize GRPC connections.
        self._init_grpc_connnections(api_key=api_key)

        self._verify_api_key()
        # if api key is an api key and not a session, set project
        if api_key.startswith("apikey/"):
            api_key_project = self._api_key_project(api_key=api_key)
            # check api key project with configured project
            if (
                self.config.project != ""
                and self.config.project != api_key_project.name
            ):
                raise Exception(
                    f"\nERROR: apikey used is not valid for current configured project [{self.config.project}]\nview config with 'continual config show'\nclear config with 'continual config clear-all'"
                )
            self.config.set_project(project=api_key_project.name, save=save)

        self.config.set_api_key(api_key=api_key, save=save)

        # set environment to default
        if hasattr(self, "_projects"):
            if not self.config.environment:
                self.set_config_environment(environment="", save=save)
        else:
            if not self.config.environment:
                self.config._environment = ""

        # Reset managers
        self._init_managers()

    def _verify_project(self, project: str) -> str:
        # Verify project name and return fully qualified name
        try:
            for p in self.projects.list_all():
                if project == p.name or project.split("/")[-1] in {
                    p.id,
                    p.display_name,
                }:
                    return p.name

            raise Exception(
                f"Project '{project}' not found. Make sure to provide fully qualified project name or unique project ID."
            )
        except Exception as e:
            raise Exception(f"Unable to verify project '{project}'. {str(e)}")

    def set_config_project(
        self, project: str, save: bool = False, environment: str = None
    ) -> str:
        """Sets config project.

        Arguments:
            project: The project to set on this client.
            save: Whether to save the project to the config file on disk.

        Examples:
            >>> from continual import Client
            >>> c = Client(verify=False)
            >>> c.config.project # If a YAML config is present in the OS env or filesystem this will not be empty
            ''
            >>> c.set_config_project(project='example-project', save=False) # Attempt to verify that project exists, API key must be valid
            >>> c.config.project
            'example-project'
        """
        # Here ensure that if the API key is blank, you cannot set a project (you need to login first)
        if not self.config.api_key and project:
            raise Exception(
                f"API key is empty. Cannot set project '{project}' until client has valid API key."
            )

        project_name = self._verify_project(project=project)
        self.config.set_project(project=project_name, save=save)
        # set environment to default env
        self.set_config_environment(environment=environment, save=save)

        # Reset managers
        self._init_managers()

    def _verify_environment(self, project: str, environment: str) -> str:
        # Verify environment name after project is verified
        try:
            proj = self.projects.get(project)
            if environment == None or environment == "":
                if proj is None:
                    return ""
                environment = proj.default_environment
            env_identifier_to_verify = ProjectEnvironmentIdentifer(
                project_name_or_id=project, environment_name_or_id=environment
            )
            environment_names = [e.name for e in proj.environments.list_all()]
            if env_identifier_to_verify.environment_name not in environment_names:
                raise Exception(f"Environment '{environment}' not found.")

            return environment
        except Exception as e:
            raise Exception(f"Unable to verify environment '{environment}'. {str(e)}")

    def set_config_environment(self, environment: str, save: bool = False) -> None:
        """Sets config environment.

        Arguments:
            environment: The environment to set on this client.
            save: Whether to save the environment to the config file on disk.

        Examples:
            >>> from continual import Client
            >>> c = Client(verify=False)
            >>> c.config.environment # If a YAML config is present in the OS env or filesystem this will not be empty
            ''
            >>> c.set_config_environment(environment='example-env', save=False) # Attempts to verify that environment exists, API key and project must be valid
            >>> c.config.environment
            'example-env'
        """

        environment = self._verify_environment(
            project=self.config.project, environment=environment
        )
        self.config.set_environment(environment=environment, save=save)

        # Reset managers
        self._init_managers()

    def __del__(self) -> None:
        if hasattr(self, "_mgmt_channel"):
            self._mgmt_channel.close()
        if hasattr(self, "_fs_channel"):
            self._fs_channel.close()
        if hasattr(self, "_gw_channel"):
            self._gw_channel.close()

    def viewer(self) -> User:
        """Currently authenticated user."""
        resp = self._management.GetViewer(Empty())
        return User.from_proto(resp, client=self)

    def check_viewer(self) -> None:
        """Verify if client's user is authenticated."""
        self._management.CheckViewer(Empty())

    def _api_key_project(self, api_key: str) -> Project:
        """Get project associated with the API key."""
        req = management_pb2.GetApiKeyProjectRequest(name=api_key)
        resp = self._management.GetApiKeyProject(req)
        return Project.from_proto(resp, client=self)

    def register(
        self,
        first_name: str,
        last_name: str,
        email: str,
        password: str,
        save: bool = True,
    ) -> User:
        """Register a new account.

        This function registers and authenticates a new user
        and saves the authenticated user information in local config.

        Arguments:
            first_name: First name of user.
            last_name: Last name of user
            email: Email address.
            password: Password.
            save: Whether to persist API key to config file

        Returns:
            The newly created authenticated user.

        Examples:
            >>> from continual import Client
            >>> client = Client(verify=False)
            >>> client.register(first_name='test', last_name='user', email='test@continual.ai', password='test123')
            <User object {'name': 'users/zxNbTXkbxLeyjb3SUhJ2fR', 'email': 'test@continual.ai', 'email_verified': True,
            'full_name': 'test user', 'update_time': '2022-12-15T01:00:20.707583Z', 'create_time': '2022-12-15T01:00:20.707583Z',
            'trial_available': True, 'first_name': 'test', 'last_name': 'user', 'bio': '', 'location': '', 'password': '',
            'service_account': False, 'disabled': False}>
            >>> client.config.show()
            Config Files:
            ${HOME}/continual/continual.yaml
            ${HOME}/continual/.continual.yaml
            Email: test@continual.ai
            Endpoint: http://sdk.continual.ai
            Project:
            Environment:
            Style: GREEN
            Raise Exception: True
            Debug:
            API Key: *******************************
        """
        req = management_pb2.RegisterRequest(
            first_name=first_name, last_name=last_name, email=email, password=password
        )
        resp = self._management.Register(req)

        # No need to reverify API key in response
        self.config.set_api_key(resp.auth_token, save=save)
        self._init_grpc_connnections(self.config.api_key)
        self.config._email = email
        return User.from_proto(resp.user, client=self)

    def login(self, email: str, password: str, save: bool = True) -> User:
        """Login to Continual.

        It is strongly recommended to use `continual login` CLI
        or an API key instead of logging in via the SDK.

        Args:
            email: Email address.
            password: Password.
            save: Whether to persist API key to config file
        Returns:
            The authenticated user.

        Examples:
            >>> from continual import Client
            >>> client = Client(verify=False)
            >>> client.login(email='test@continual.ai', password='test123') # Assuming user is already registered
            <User object {'name': 'users/zxNbTXkbxLeyjb3SUhJ2fR', 'email': 'test@continual.ai', 'email_verified': True,
            'full_name': 'test user', 'update_time': '2022-12-15T01:00:20.707583Z', 'create_time': '2022-12-15T01:00:20.707583Z',
            'trial_available': True, 'first_name': 'test', 'last_name': 'user', 'bio': '', 'location': '', 'password': '',
            'service_account': False, 'disabled': False}>
            >>> client.config.show()
            Config Files:
            ${HOME}/continual/continual.yaml
            ${HOME}/continual/.continual.yaml
            Email: test@continual.ai
            Endpoint: http://sdk.continual.ai
            Project:
            Environment:
            Style: GREEN
            Raise Exception: True
            Debug:
            API Key: *******************************
        """
        req = management_pb2.LoginRequest(email=email, password=password)
        resp = self._management.Login(req)

        # No need to reverify API key in response
        self.config.set_api_key(resp.auth_token, save=save)
        self._init_grpc_connnections(self.config.api_key)
        self.config._email = email
        return User.from_proto(resp.user, client=self)

    def logout(self) -> None:
        """Logout.

        Logs out current session deleting the associated API key.
        """
        if self.config.api_key is not None and self.config.api_key != "":
            self._management.Logout(Empty())
            self.config.set_api_key(api_key=None)
            self.config._email = None
            self.config.save()

    def version(self) -> Tuple[str, str, bool]:
        """Get the current server version.

        Returns:
          A tuple of the client version string, server version string, and a boolean
          true if client needs an update.
        """
        client_version = __version__
        req = management_pb2.GetServerVersionRequest(client_version=client_version)
        resp = self._management.GetServerVersion(req)
        return client_version, resp.server_version, resp.upgrade_required
